// -*- C++ -*-
/***************************************************************************
                          kapitalist.h  -  description
                             -------------------
    begin                : Mon Jun 11 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef KAPITALIST_H
#define KAPITALIST_H
 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

//
// Qt / KDE
//

#include <qtimer.h>
#include <qsize.h>

class QSocketNotifier;

#include <kapp.h>
#include <kmainwindow.h>
#include <kaccel.h>
#include <kaction.h>

//
// App
//

#include "global.h"
#include "packets.h"

class ConnectDialog;
class HousesDialog;
class TradeDialog;
class ConfigDialog;

#define BUF_SIZE	255


//
// Game
//

#include "actions.h"

class KapGame;


//
// View
//

class KapView;

class InputLine;


// to store (packet type, packet *) pairs
struct PacketQueueEntry
{
  int type;
  void *packet;
};

typedef QList<PacketQueueEntry> PacketQueue;



/**
  * The base class for Kapitalist application windows. It sets up the
  * main window and reads the config file as well as providing a
  * menubar, toolbar and statusbar. An instance of KapView
  * creates your center view, which is connected to the window's KapGame
  * object.  KapitalistApp reimplements the methods that KMainWindow
  * provides for main window handling and supports full session
  * management as well as using KActions.  
  * 
  * @see KMainWindow 
  * @see KApplication
  * @see KConfig
  */
class KapitalistApp : public KMainWindow
{
  Q_OBJECT

  friend class KapView;
  friend class InputLine;
  friend class ConnectDialog;

  // the following is needed for signal emission (e.g. for animations)
  friend class PlayerView;
  friend class GameInfo;

public:

  /**
   * constructor of KapitalistApp, calls all init functions to create
   * the application.
   */
  KapitalistApp(QWidget* parent=0, const char* name=0);
  /** 
   * Destructor of KapitalistApp
   */
  ~KapitalistApp();
  /**
   * returns a pointer to the game instance connected to this KapitalistApp
   * window
   */
  KapGame *game() const { return _game; } 	
  /**
   * returns a pointer to the view instance of this KapitalistApp window
   */
  KapView *view() const { return _view; }

  /**
   * returns a pointer to the connection instance of this KapitalistApp window
   */
  struct connection *conn() const { return const_cast<connection *>(&_conn); }

  /**
   * returns the current client state which can be one of the states
   * defined in common/global.h
   */
  enum client_states const client_state() { return _client_state; }
  
  /**
   * creates and connects the socket notifier, which emits a signal
   * when data comes from the server
   */
  void open_sockNotifier();

protected:

  /**
   * queue for storing the packets before they are processed one by
   * one. It contains PacketQueueEntry objects.
   */
  PacketQueue _packetQueue;

  /**
   * indicates some animation is in progress right now.
   */
  bool _animationInProgress;
  
protected:

  /**
   * save general options like all bar positions and status as well
   * as the geometry to the configuration file
   */
  void saveOptions();

  /**
   * read general options again and initialize all variables
   */
  void readOptions();

  /**
   * initializes the KActions of the application
   */
  void initActions();

  /**
   * sets up the statusbar for the main window by initialzing a
   * statuslabel.
   */
  void initStatusBar();

  /**
   * initializes the game object of the main window that is
   * connected to the view in initView().
   *
   * @see initView();
   */
  void initGame();

  /**
   * creates the centerwidget of the KapitalistApp window and sets
   * it as the view
   */
  void initView();

  /**
   * queryClose is called by KTMainWindow on each closeEvent of a
   * window. Against the default implementation (only returns true),
   * this calls saveModified() on the game object to ask if the
   * game shall be saved if Modified; on cancel the closeEvent is
   * rejected.
   *
   * @see KMainWindow#queryClose
   * @see KMainWindow#closeEvent 
   */
  virtual bool queryClose();

  /**
   * queryExit is called by KTMainWindow when the last window of the
   * application is going to be closed during the closeEvent().
   * Against the default implementation that just returns true, this
   * calls saveOptions() to save the settings of the last window's
   * properties.
   *
   * @see KMainWindow#queryExit
   * @see KMainWindow#closeEvent
   * @see saveOptions();
   */
  virtual bool queryExit();

  /**
   * saves the window properties for each open window during session
   * end to the session config file, including saving the currently
   * open games by a temporary filename provided by KApplication.
   * 
   * (TODO)
   *
   * @see KMainWindow#saveProperties
   */
  virtual void saveProperties(KConfig *_cfg);

  /**
   * reads the session config file and restores the application's
   * state including the last open games by reading the temporary
   * files saved by saveProperties()
   *
   * (TODO)
   *
   * @see KMainWindow#readProperties
   */
  virtual void readProperties(KConfig *_cfg);
  
  
  /**
   * handles packets when they are processed (i.e. when the packet
   * queue is processed)
   */
  void handle_packet_input(char *packet, int type);
  
  /**
   * queues the given packet for later processing 
   */
  void queue_packet_input(void *packet, int type);


  ////////////////////////////////////////////////////////////////////
  //  individual packet handlers (general issues, not game stuff!)  //
  ////////////////////////////////////////////////////////////////////

  /**
   * handles JOIN_GAME_REPLY packets (reply from server after the
   * client has requested to join a game). The answer can be positive
   * or negative.
   */
  void handle_join_game_reply(struct packet_join_game_reply *packet);
  
  /**
   * handles CHAT messages (which can also be error messages by the server)
   */
  void handle_chat_msg(struct packet_generic_message *packet);

  /**
   * handles GAME_STATE packets (by which the server indicates what's
   * the current state of the game, e.g. game running, game over, etc.)
   */
  void handle_game_state(struct packet_generic_integer *packet);
  
  /**
   * handles CONFIRM_ANIM_FINISHED packets (server orders the client
   * to send a confirmation when it has finished all animations)
   */
  void handle_confirm_anim_finished(struct packet_generic_integer *packet);


  //////////////////////////////////////////////////////
  //  send methods (general issues, not game stuff!)  //
  //////////////////////////////////////////////////////
  
  /**
   * sends a confirmation that client has finished all animations
   */
  void send_anim_finished();



  /**
   * sets the client state to "newstate" (see common/global.h for a
   * list of the possible client states)
   */
  void set_client_state(enum client_states newstate);
  

  //////////////////////////////////////////////////////
  //                  network methods                 //
  //////////////////////////////////////////////////////

  /**
   * forks a server and connects to it. Returns FALSE if we find we
   * couldn't start the server or couldn't connect.
   */
  bool client_start_server_and_connect(const char *name, const int port);

  /**
   * connects to the server running at "hostname" on port "port" and use
   * "name" as nickname.
   */
  int connect_to_server(const char *name, const char *hostname, 
			const int port,
			char *errbuf, size_t errbufsize);
  /**
   * close the connection to the server (implies a client state change)
   */
  void disconnect_from_server(void);

//    void get_net_input(int *fid);


  //////////////////////////////////////////////
  //                animations                //
  //////////////////////////////////////////////

  /**
   * sets a bool that animations have started
   */
  void animationStarted();
  /**
   * resets the bool (animations have finished)
   */
  void animationFinished();


  ////////////////////////////////////////////////
  //                action stuff                //
  ////////////////////////////////////////////////

  /**
   * toggles the actions given by flags (which is the binary and of
   * the actionflags listed in actions.h)
   */
  void enable_actions(int flags, bool enable=true);

  /**
   * disables the actions given by flag (shortcut to enable_actions(...,false)
   */
  void disable_actions(int flags);

public slots:

  /**
   * opens a new application window by creating a new instance of
   * KapitalistApp
   */
  void slotGameNewWindow();
  /**
   * connects to the server
   */
  void slotGameConnect();
  /**
   * disconnects from server
   */
  void slotGameDisconnect();
  /**
   * starts the game on the server (if allowed)
   */
  void slotGameStart();

  /**
   * asks for saving if the game is modified (TODO), then closes the
   * actual game and window
   */
  void slotGameClose();

  /**
   * closes all open windows by calling close() on each memberList
   * item until the list is empty, then quits the application.  If
   * queryClose() returns false because the user canceled the
   * saveModified() dialog, the closing breaks (TODO).
   */
  void slotGameQuit();

  /**
   * executes slotRollDice
   */
  void slotActionRollDice();
  /**
   * executes slotActionDone
   */
  void slotActionDone();
  /**
   * executes slotActionPropertyBuy with the location of the player's token
   */
  void slotActionPropertyBuy();
  /**
   * executes slotPayDebts
   */
  void slotActionPay();
  /**
   * executes slotBankruptcy
   */
  void slotActionBankruptcy();

  /**
   * opens a jail dialog
   *
   * @see JailDialog
   */
  void slotActionJail();
  /**
   * opens the house dialog for a given monopoly
   *
   * @see HousesDialog
   */
  void slotActionHouses(int monopid);
  /**
   * opens the trade dialog
   *
   * @see TradeDialog
   */
  void slotActionTrade();

  /**
   * toggles the toolbar
   */
  void slotViewToolBars();
  /**
   * toggles the statusbar
   */
  void slotViewStatusBar();
  /**
   * opens the configure dialog
   *
   * @see TradeDialog
   */
  void slotSettingsConfigure();


  /**
   * changes the statusbar contents for the standard label
   * permanently, used to indicate current actions.
   * 
   * @param text the text that is displayed in the statusbar
   */
  void slotStatusMsg(const QString &text);
  
  
  /////////////////////////////////
  //    network/packet handling  //
  /////////////////////////////////

  /**
   * reads input from socket (when available). This slot is connected
   * to the socket notifier
   */
  void slotNetInput();

  /**
   * processes the packet queue until the queue is empty. When there
   * are no packets, idle cycles are performed
   */
  void slotProcessPacketQueue();


  ///////////////////////////////////
  //    requests sent to server    //
  ///////////////////////////////////

  /**
   * sends roll request to server
   */
  void slotRollDice(int dicesum=-1);

  /**
   * sends done request to server
   */
  void slotDone();

  /**
   * sends pay request to server
   */
  void slotPayDebts();

  /**
   * sends bankruptcy request to server
   */
  void slotBankruptcy();


  /**
   * send property buy request to server
   */
  void slotPropertyBuy(int property=-1);

  /**
   * sends property mortgage request to server
   */
  void slotPropertyMortgage(int property);

  /**
   * send property unmortgage request to server
   */ 
  void slotPropertyUnmortgage(int property);

//    void slotHousesBuy(int monopid, int nhouses[MAX_STREETS_IN_MONOP]);
//    void slotHousesSell(int monopid, int nhouses[MAX_STREETS_IN_MONOP]);

  /**
   * send jail pay request to server
   */
  void slotJailPay();
  /**
   * send jail use card request to server
   */
  void slotJailCard();
  
//    void updateHousesDialog(struct packet_monop_info *) const;
//    void updateTradeDialog() const;

signals:

  /**
   * emitted when a new packet has been queued
   */
  void packetQueued();

//    void packetQueueEntryFinished();

private:

  /**
   * the configuration object of the application
   */
  KConfig *config;
  /**
   * represents your actual game (more or less mirrored data from
   * server) and is created only once per KapitalistApp instance.
   */
  KapGame *_game;
  /**
   * the main widget which represents the central window area and is
   * the parent of all elements (e.g. boardview, etc.)
   */
  KapView *_view;
  
  /**
   * contains the preferred size read from the users config file
   */
  QSize _preferredSize;


  ///////////////////////////////////////////
  //                actions                //
  ///////////////////////////////////////////

  /**
   * general actions (connecting to server, new windows, etc.),
   * contained in the Game menu
   */
  KAction* gameNewWindow;
  KAction* gameConnect;
  KAction* gameDisconnect;
  KAction* gameStart;
  KAction* gameClose;
  KAction* gameQuit;

  /**
   * game control actions (roll, done, bankruptcy, etc.),
   * contained in the Actions menu
   */
  KAction* actionRollDice;
  KAction* actionDone;
  KAction* actionPropertyBuy;
  KAction* actionPay;
  KAction* actionBankruptcy;

  /**
   * dialog actions, also contained in the Actions menu
   */
  KAction* actionJail;
//    KAction* actionProperty;
//    KAction* actionHouses;
  KAction* actionTrade;

  /**
   *  setting actions, contained in the Settings menu
   */
  KToggleAction* viewToolBars;
  KToggleAction* viewStatusBar;
  
  KAction *settingsConfigure;

  /**
   * modal dialog for connecting to the server enter connection data
   * (entering name, server hostname, port)
   *
   * @see ConnectDialog
   */
  ConnectDialog *_connectDlg;

//    HousesDialog *_housesDlg[BOARD_N_MONOPS];

  /**
   * trade dialog (containing all trades to other players)
   *
   * @see TradeDialog
   */
  TradeDialog *_tradeDlg;

  /**
   * config dialog
   *
   * @see TradeDialog
   */
  ConfigDialog *_configDlg;


  /**
   * the current client state (see common/global.h for possible client states)
   */
  enum client_states _client_state;
  

  //////////////////////////////
  //    network attributes    //
  //////////////////////////////

  /**
   * the client's connection to the server 
   */
  struct connection _conn;

  /**
   * the socket notifier which listens to the connection and activates
   * slotNetInput in case of data to be read.
   *
   * @see QSocketNotifier
   */
  QSocketNotifier *_sockNotifier;

};
 
#endif // KAPITALIST_H


