// -*- C++ -*-
/***************************************************************************
                          kapview.cpp  -  description
                             -------------------
    begin                : Mon Jun 11 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//
// System
//

#include <cstdio>
#include <cmath>

//
// Qt / KDE
//

#include <qprinter.h>
#include <qpainter.h>

#include <qsocketnotifier.h>
#include <qlayout.h>
#include <qtabwidget.h>
#include <qvbox.h>

#include <qcanvas.h>

#include <qstring.h>
#include <qregexp.h>

#include <kstddirs.h>
#include <klocale.h>

//
// App
//

#include "global.h"
#include "version.h"
#include "events.h"
#include "kapitalist.h"

//
// Game
//

#include "kapgame.h"
#include "player.h"
#include "board.h"
#include "field.h"
#include "property.h"
#include "street.h"

//
// View
//

#include "kapview.h"
#include "centerview.h"
#include "leftview.h"
#include "playerview.h"
#include "boardview.h"
#include "fieldview.h"
#include "cardview.h"
#include "diceview.h"

#include "gameinfo.h"
#include "myplrinfo.h"
#include "myholdingsinfo.h"
#include "playerinfo.h"

#include "inputline.h"
#include "outputwindow.h"

#include "themedata.h"

// to ensure that the packets come in the right order
static bool plr_info_received = false;

// message in gameinfo for first street on which houses are bought
// should be set, not appended (is set in handle_field_info and unset in 
// handle_monop_info (which comes after all field infos in E_BUY_HOUSES bundle)

static bool append_msg_houses_event = false;

static struct monop_locale_struct monop_locale[BOARD_N_MONOPS] = 
#include "monops.us.locale"
;

static struct card_locale_struct cchest_card_locale[GAME_N_COMCHESTS] =
#include "cchest_cards.us.locale"
;				 
static struct card_locale_struct chance_card_locale[GAME_N_CHANCES] =
#include "chance_cards.us.locale"
;				 


#define DELETE(x)	(if (x) delete x; x = 0;)

KapView::KapView(QWidget *parent, const char *name) 
  : QWidget(parent, name),
    _top(0), 
    _leftFrm(0), 
    _left(0), 
//      _leftColor(QColor(150,150,150)),
//      _leftColor(QColor(0,170,255)),
    _leftColor(QColor("#707070")),
    _board(0),
    _centerFrm(0),
    _center(0), 
//      _centerColor(FieldView::baseColor().dark(130)),
    _centerColor(QColor("#707070")),
    _logoLab(0),
    _centerTop(0),
    _centerMid(0),
    _centerBot(0),
    _textIOTab(0),
    _centerCanv(0),
    _gameinfo(0),
    _chatIOFrm(0),
    _chatIO(0),
    _chatHistory(0),
    _chatIn(0),
    _chatOut(0),
    _gameIOFrm(0),
    _gameIO(0),
    _gameCmdHistory(0),
    _gameCmdIn(0),
    _gameOut(0),
//      _card(0),
//      _dice(0),
    _myinfo(0),
    _myhold(0),
    _plrinfo(0),
    _idleTimer(this),
    _finishIdleCycle(false)
{
  int i;

//    preloadPixmaps();

  // load default theme
  _theme = new ThemeData();

  for (i=0; i<GAME_MAX_PLAYERS; ++i)
    _plr[i] = 0;

//    locales_eliminate_hyphens();

  setBackgroundMode(PaletteBase);

  _board = new BoardView(this, this);
  _centerFrm = new CenterView(this, _board);
  _board->add_center(_centerFrm);

  _textIOTab = new QTabWidget(_centerFrm);
  _textIOTab->setMinimumSize( QSize(300,150) );
//    _textIOTab->setPalette( palette() );

  _chatIOFrm = new QFrame(_textIOTab);
  _chatIO = new QVBoxLayout(_chatIOFrm);

  _chatOut = new OutputWindow(this, _chatIOFrm, OutputWindow::CHAT_MODE);
  _chatHistory = new UpDownFrame(this, _chatIOFrm);
  _chatIn = new InputLine(this, _chatHistory, InputLine::CHAT_MODE);
  _chatHistory->set_inputLine(_chatIn);
  _chatIn->set_frame(_chatHistory);

  _chatIO->addWidget(_chatOut, 1);
  _chatIO->addWidget(_chatHistory, 0);

  _textIOTab->addTab(_chatIOFrm, i18n("&Chat"));
  
  createEmptyViews();

  connect( &_idleTimer, SIGNAL(timeout()), SLOT(slotIdleCycle()) );

}

KapView::~KapView()
{
  // most widgets are children of this, so no explicit deletion required
  delete _theme;
}

KapitalistApp *KapView::getApp() const
{
  KapitalistApp *theApp= dynamic_cast<KapitalistApp *>(parentWidget());
  assert(theApp);
  return theApp;
}

PlayerView * const KapView::myPlrView() const
{
  return _plr[getGame()->myPID()];
}

PlayerView * const KapView::plr(int id) const
{
  assert(!getGame()->nobody(id));
  return _plr[id];
}

PlayerView * const KapView::currPlr() const
{
  Player *currplr = getGame()->currPlr();
  if (currplr == 0) return 0;
  else return _plr[getGame()->currPlr()->id()];
}

QPixmap KapView::getPixmap(const QString &name) const
{
  return _theme->getPixmap(name);
}

QPixmap KapView::getPixmap(const int fieldID) const
{
  return _theme->getPixmap(fieldID);
}

QBitmap KapView::getBitmap(const QString &name) const
{
  return _theme->getBitmap(name);
}

QPixmap KapView::getDiePixmap(const int i, const int j) const
{
  return _theme->getDiePixmap(i, j);
}

QCanvasPixmapArray *KapView::dice() const
{
  return _theme->dice();
}

void KapView::add_left(QWidget *newWidget, int stretch)
{
  _left->addWidget(newWidget, stretch);
//    newWidget->setBackgroundColor(_leftColor);
}

void KapView::add_center(QWidget *newWidget, int stretch)
{
  _center->addWidget(newWidget, stretch);
//    newWidget->setBackgroundColor(_centerColor);
}

void KapView::add_center(QLayout *newLayout, int stretch)
{
  _center->addLayout(newLayout, stretch);
}

void KapView::createViews()
{
  int i;

  // recreate layouts
  // (deleting _center deletes all sublayouts too)
  if (_center) delete _center; _center = 0;
  if (_left) delete _left; _left = 0;
  if (_top) delete _top; _top = 0;

  // TODO: delete infos on left frame and center frame

  _top = new QGridLayout(this, 2,3);
  _top->setRowStretch(0,1);
  _top->setColStretch(0,0);	// left
  _top->addColSpacing(0,250);
  _top->setColStretch(1,10);	// board

  // *******  left area    ********

  _leftFrm = new LeftView(this, this);
  _top->addWidget(_leftFrm, 0,0);
  _top->addWidget(_board, 0,1);

  // create layout for left frame
  _left = new QVBoxLayout(_leftFrm, 5);
  
  _center = new QVBoxLayout(_centerFrm, 2);
  _center->addSpacing(10);

  //
  // game controls (in center and left frame)
  //
  
  // TODO: assert all elements which survive at state change
  assert(_board);
  assert(_textIOTab);
  

  // **************** from here ****************

//    // layout for game info (centerTop)
//    _centerTop = new QGridLayout(2,3, 5);
//    add_center(_centerTop, 3);

//    _centerTop->setColStretch(0,1);
//    _centerTop->setColStretch(1,7);
//    _centerTop->setColStretch(2,1);
//    _centerTop->setRowStretch(0,1);

//    _gameinfo = new GameInfo(this, _centerFrm);
//    _centerTop->addWidget(_gameinfo, 0,1);
  
//    // layout for chance and comchest cards (central)
//    _centerMid = new QGridLayout(3,2);
//    _centerMid->addRowSpacing(1,80);  // for logo
//    add_center(_centerMid, 2);

//    _dice = new DiceView(this, _centerFrm);
//  //    _centerMid->addWidget(_dice, 1,0);
//    _dice->init_dice();

//    _card = new CardView(this, _centerFrm);
//    _centerMid->addWidget(_card, 1,1, AlignCenter);

//    // let the center column of _centerMid be of stretch 0 with defined spacing
//    _centerMid->setColStretch(0, 1);
//    _centerMid->setColStretch(1, 2);
//    _centerMid->addColSpacing(1, 6 * FieldView::DEFAULT_WIDTH); // _card->sizeHint().width()

  // **************** to here ****************

  // **************** temporary start ****************

//    _gameinfo = new GameInfo(this, _centerFrm);
//    _gameinfo->hide();

//    _dice = new DiceView(this, _centerFrm);
//    _dice->init_dice();
//    _dice->hide();

//    _card = new CardView(this, _centerFrm);
//    _card->hide();

  // **************** temporary end ****************

  // create and setup canvas and view it with game info
  _centerCanv = new QCanvas(_centerFrm);
  _centerCanv->setBackgroundColor(_centerColor);
  _centerCanv->setAdvancePeriod(-1);
  _gameinfo = new GameInfo(this, _centerCanv, _centerFrm);
//    add_center(_centerCanvView, 4);


//    QCanvasSprite* cvChanceCard 
//      = new QCanvasSprite(new QCanvasPixmapArray("kaptest.png"),
//  			_centerCanv);
//    cvChanceCard->setAnimated(false);
//    cvChanceCard->move(10,20);
//    cvChanceCard->show();

  // layout for canvas
  _centerTop = new QHBoxLayout();
  add_center(_centerTop, 6);
  
  _centerTop->addSpacing(10);
  _centerTop->addWidget(_gameinfo, 1);
  _centerTop->addSpacing(10);
  
  // layout for text IO (central)
  _centerBot = new QHBoxLayout();
  add_center(_centerBot, 4);
  
  _centerBot->addSpacing(10);
  _centerBot->addWidget(_textIOTab, 1);
  _centerBot->addSpacing(10);
  
  // populate _leftFrm

  _myinfo = new MyPlrInfo(this, _leftFrm);
  _myhold = new MyHoldingsInfo(this, _leftFrm);
  _plrinfo = new PlayerInfo(this, _leftFrm);

  add_left(_myinfo, 0);
  add_left(_myhold, 0);
  add_left(_plrinfo, 0);


  for (i=0; i<GAME_MAX_PLAYERS; ++i)
    _plr[i] = 0;

  _gameIOFrm = new QFrame(_textIOTab);
  _gameIO = new QVBoxLayout(_gameIOFrm);

  _gameOut = new OutputWindow(this, _gameIOFrm, OutputWindow::GAME_MODE);
  _gameCmdHistory = new UpDownFrame(this, _gameIOFrm);
  _gameCmdIn = new InputLine(this, _gameCmdHistory, InputLine::GAME_MODE);
  _gameCmdHistory->set_inputLine(_gameCmdIn);
  _gameCmdIn->set_frame(_gameCmdHistory);

  _gameIO->addWidget(_gameOut, 1);
  _gameIO->addWidget(_gameCmdHistory, 0);

  _textIOTab->addTab(_gameIOFrm, i18n("&Game Log"));

  _center->addSpacing(10);

  _left->addStretch(1);

  // logo moves to center
  _centerFrm->set_logo_alignment(AlignCenter);

  // activate layouts
  _left->activate();
  _center->activate();

  _top->activate();

  setMinimumSize( QSize(760,560-60) );

//    adjustSize();

  show();
}

void KapView::createEmptyViews()
{
  // recreate layouts
  // (deleting _center deletes all sublayouts too)
  if (_center) delete _center; _center = 0;
  if (_left) delete _left; _left = 0;
  if (_top) delete _top; _top = 0;

//    _board->delete_board();
//    if (_gameinfo) delete _gameinfo; _gameinfo = 0;
//    if (_dice) delete _dice; _dice = 0;
//    if (_card) delete _card; _card = 0;
//    if (_myinfo) delete _myinfo; _myinfo = 0;
//    if (_myhold) delete _myhold; _myhold = 0;
//    if (_plrinfo) delete _plrinfo; _plrinfo = 0;
  
    // layouts:
  //	top:	grid 1 row, 2 cols
  //		1st col contains holdings, budget, player infos
  //		2nd col contains board and game infos + inputline in the center

  _top = new QGridLayout(this, 2,3);
  _top->setRowStretch(0,1);
  _top->setColStretch(0,0);	// left
  _top->setColStretch(1,10);	// board

  // *******  left area    ********
  // don't show left area
//    _top->addWidget(_leftFrm, 0,0);


  // *******  board is on the right   *********
  _top->addWidget(_board, 0,1);

//    _left = new QVBoxLayout(_leftFrm, 5);
//    _left->addStrut(200);
//    QLabel *_leftlogo = new QLabel(_leftFrm);
//    _leftlogo->setPixmap(getPixmap("info"));
//  //    _leftlogo->setText("Left Logo");
//    _leftlogo->setMinimumSize( _leftlogo->sizeHint() );
//    _leftlogo->show();
//    _left->addWidget(_leftlogo);
//    _left->addStretch(1);

//    _left->addWidget(_leftlogo);

  // *******  central space (center of board)    ********

  _center = new QVBoxLayout(_centerFrm, 2);
  _center->addStretch(2);

  assert(_textIOTab);

  // layout for text IO (central)
  _centerBot = new QHBoxLayout();
  add_center(_centerBot, 4);
  
  _centerBot->addSpacing(10);
  _centerBot->addWidget(_textIOTab, 1);
  _centerBot->addSpacing(10);

  _center->addStretch(1);

  // *******  end central space (center of board)    ********


  _center->activate();
//    _left->activate();

  _top->activate();

  _centerFrm->set_logo_alignment(AlignTop);

  _centerFrm->adjustSize();
//    _leftFrm->adjustSize();

//    _board->add_center(_centerFrm);
  _board->adjustSize();

  setMinimumSize( sizeHint() );

  adjustSize();

  show();
}


void KapView::changeTheme(const QString& namesFname, const QString& boardFname,
			  const QString& diceFname)
{
  if (_theme) delete _theme;
  _theme = new ThemeData(namesFname, boardFname, diceFname);
  // refresh the board
  _board->rereadThemeData();
  _myinfo->update();
  _plrinfo->update();
}


void KapView::appendChatMsg(const char *msg) const
{ 
  _chatOut->append(replace_locales(msg)); 
}

void KapView::appendGameMsg(const char *msg) const
{ 
  _gameOut->append(replace_locales(msg)); 
}

void KapView::appendCurrOut(const char *msg) const
{
  if (_textIOTab->currentPage() == _chatIOFrm) {
    appendChatMsg(msg);
  } else if (_textIOTab->currentPage() == _gameIOFrm) {
    appendGameMsg(msg);
  }
}

QString KapView::replace_locales(const QString &text) const
{
  QString newtext = text;

  int pos = 0;

  // generate the regular expression matching strings like "%14"
  char rxchar[10];

  // process field names
  sprintf(rxchar, "(\\%c\\d+)", CHAT_FIELDNAME_PREFIX);
  QRegExp rx_fieldname(rxchar);

  while ( pos >=0 ) {
    // find next occurence of the regular expression (if there is no more,
    // pos becomes -1 and the loop is exit)
    pos = rx_fieldname.search(newtext, pos);
    if (pos >= 0) {
      pos++;
      // extract the substring matched by the regular expression
      QString locstr = newtext.mid(pos,rx_fieldname.matchedLength()-1);
      int loc = locstr.toInt();
      if (loc>=0 && loc<BOARD_N_FIELDS)
	newtext.replace(QChar(CHAT_FIELDNAME_PREFIX) + locstr,
			field_locale(loc)->name);
    }
  }

  // process money
  sprintf(rxchar, "(\\%c\\d+)", CHAT_MONEY_PREFIX);
  QRegExp rx_money(rxchar);

  pos = 0;

  while ( pos >=0 ) {
    // find next occurence of the regular expression (if there is no more,
    // pos becomes -1 and the loop is exit)
    pos = rx_money.search(newtext, pos);
    if (pos >= 0) {
      pos++;
      // extract the substring matched by the regular expression
      QString moneystr = newtext.mid(pos,rx_money.matchedLength()-1);
      int money = moneystr.toInt();
      if (money>=0) {
	QString newmoneystr = money_locale(money);
	newtext.replace(QChar(CHAT_MONEY_PREFIX) + moneystr, newmoneystr);
      }
    }
  }

  return newtext;
}


void KapView::handle_your_turn(struct packet_generic_integer *packet)
{
  // who is acting ?
  int plrid = packet->value;
  bool me = (plrid == getGame()->myPID());
  const char *name = (me)? "you" : getGame()->player(plrid)->name();
  const char *ucname = (me)? "You" : name;
  const char *jailturnstr[5] = { "0th", "1st", "2nd", "3rd", "4th" }; 

  assert(!getGame()->nobody(plrid));	// should be a valid plrid

  Player *plr = getGame()->player(plrid);

  char msg[MSG_SIZE];

  char head[MSG_SIZE];
  char details[MSG_SIZE];
  
  // first clear everything not necessary any more (i.e. cards)
  _gameinfo->hide_card();
//    _card->hide();

  sprintf(msg, "%s%s turn", ucname, (me)?"r":"'s");
  sprintf(head, msg);
//    sprintf(head, "%s% dice%s", ucname, (me)?"":"s");
  
  if(plr->location() == JAIL) {
    int ninjail = plr->nturns_in_jail();
    sprintf(msg+strlen(msg), " (%s turn in jail)", 
	    jailturnstr[ninjail]);
    sprintf(details, "It's %s%s %s turn in jail.", name, (me)?"r":"'s",
	    jailturnstr[ninjail]);
  } else {	// not in jail
    sprintf(msg+strlen(msg), " (on field %s (%d))",
    	    field_locale(plr->location())->name, plr->location());
//      sprintf(details, "%s %s on field %s (%d).", ucname, (me)?"are":"is",
//    	    field_locale(plr->location())->name, plr->location());
    *details = '\0';
  }
  
  appendGameMsg(msg);
  _gameinfo->message(MSG_INFO, head, details);

  if (me) {
    getApp()->enable_actions(A_ROLL_DICE | A_PAY | A_BANKRUPTCY |
			     A_PROPERTY_BUY |			     
			     A_JAIL | A_PROPERTY | A_HOUSES);
    getApp()->disable_actions(A_DONE);
  } else { // paranoia
    getApp()->disable_actions(A_ROLL_DICE | A_DONE | A_PAY | A_BANKRUPTCY | 
			      A_PROPERTY_BUY | 
			      A_JAIL | A_PROPERTY | A_HOUSES);
  }

}

void KapView::handle_dice_rolled(struct packet_dice_rolled *packet)
{
  // who is acting ?
  int plrid = packet->playerid;
  bool me = (plrid == getGame()->myPID());
  const char *name = (me)? "you" : getGame()->player(plrid)->name();
  const char *ucname = (me)? "You" : name;

  char msg[MSG_SIZE];
  char head[MSG_SIZE];

  sprintf(msg, "%s rolled %d and %d.", 
	  ucname, packet->dice1, packet->dice2);

  // only set header here, details follow in handle_player_info (e.g. E_MOVE)
  sprintf(head, msg);

  appendGameMsg(msg);
  _gameinfo->message(MSG_INFO, head, "");

//    _dice->set_dice(packet->dice1, packet->dice2);
  _gameinfo->set_dice(packet->dice1, packet->dice2);

  if (me) {
    getApp()->disable_actions(A_ROLL_DICE);
    getApp()->enable_actions(A_DONE);
  }
}

void KapView::handle_card_drawn(struct packet_generic_values *packet)
{
  // who is acting ?
  int plrid = packet->id;
  bool me = (plrid == getGame()->myPID());
  const char *name = (me)? "you" : getGame()->player(plrid)->name();
  const char *ucname = (me)? "You" : name;

  // which card was drawn ?
  int type = packet->value1;
  int card_no = packet->value2;

  char msg[MSG_SIZE];
  char details[MSG_SIZE];
  char head[MSG_SIZE];

  switch (type) {
    
  case COMCHEST:
  case CHANCE:
    sprintf(msg, "%s draw%s a %s card.", ucname, (me)?"":"s",
	    (type==COMCHEST)?"community chest":"chance");
    sprintf(details, msg);	// don't display the card's text

    // TODO: write card's text into gamelog
//      sprintf(msg+strlen(msg), "\n\n%s", 
//  	    (type==COMCHEST)? cchest_card_locale(card_no).msg
//  	    : chance_card_locale(card_no).msg);
    
    appendGameMsg(msg);

    _gameinfo->append_details(details);

    sprintf(head, "%s drew a %s card", ucname, 
	    (type==COMCHEST)?"community chest":"chance");
    _gameinfo->set_headline(head);
    _gameinfo->set_details(MSG_INFO, "");
    _gameinfo->draw_card(type, card_no);
    // further actions are displayed in details, when appropriate packages
    // are read
    return;
    
  default:
    assert(0);	// bug: there are no other types for cards, are there???

  }

}

void KapView::handle_trade_info(struct packet_trade *packet)
{
  int i;
//    int tradeid = packet->tradeid;
  int event = packet->event;

  // who trades with whom
  int srcid = packet->srcid;
  int destid = packet->destid;
  bool mesrc = (srcid == getGame()->myPID());
  bool medest = (destid == getGame()->myPID());
  const char *namesrc = (mesrc)? "you" : getGame()->player(srcid)->name();
  const char *namedest = (medest)? "you" : getGame()->player(destid)->name();

//    bool from_src = (packet->offer_from == T_SRC);
  int money = packet->money;

  char msg[MSG_SIZE];
  char srcmon[20], destmon[20];
  srcmon[0] = destmon[0] = '\0';

  int nprops = 0;	// count number of properties

  switch (event) {

  case E_TRADE_FINISHED:
    {
      sprintf(msg, "trade between %s and %s finished", namesrc, namedest);
      appendGameMsg(msg);
      appendChatMsg(msg);
      _gameinfo->set_headline(msg);
      _gameinfo->set_details(MSG_INFO, "");

      // money
      if (money<0) {	// src got money
	sprintf(srcmon, "$%d ", -money);
      }
      if (money>0) {	// dest got money
	sprintf(destmon, "$%d ", money);
      }

      // src

      sprintf(msg, "  %s got %s%sthe following properties:",
	      namesrc, srcmon, (*srcmon)?"and ":"");
      appendGameMsg(msg);
      _gameinfo->append_details(msg);

      // properties
      nprops = 0;
      for (i=0; i<BOARD_N_FIELDS; ++i) {
	if (packet->prop[i] == T_DEST) {
	  sprintf(msg, "    %s", field_locale(i)->name);
	  appendGameMsg(msg);
	  _gameinfo->append_details(msg);
	  nprops++;
	}
      }
      if (nprops==0) {
	sprintf(msg, "    <no properties>");
	appendGameMsg(msg);
	_gameinfo->append_details(msg);
      }

      //dest

      sprintf(msg, "  %s got %s%sthe following properties:",
	      namedest, destmon, (*destmon)?"and ":"");
      appendGameMsg(msg);
      _gameinfo->append_details(msg);

      // properties
      nprops = 0;
      for (i=0; i<BOARD_N_FIELDS; ++i) {
	if (packet->prop[i] == T_SRC) {
	  sprintf(msg, "    %s", field_locale(i)->name);
	  appendGameMsg(msg);
	  _gameinfo->append_details(msg);
	  nprops++;
	}
      }
      if (nprops==0) {
	sprintf(msg, "    <no properties>");
	appendGameMsg(msg);
	_gameinfo->append_details(msg);
      } 

    }
    return;

  case E_TRADE_OFFER:
  case E_TRADE_REPLY:
    {
      // who has proposed the offer/reply ?
      int from = packet->offer_from;
      int to = (from==T_SRC)?T_DEST:T_SRC;	// the other
      const char *namefrom = (from==T_SRC)?namesrc:namedest;
      const char *nameto = (from==T_SRC)?namedest:namesrc;
      char *moneyfrom = (from==T_SRC)?destmon:srcmon;
      char *moneyto = (from==T_SRC)?srcmon:destmon;

      // is the proposal from me ?
      bool mefrom = ( (mesrc && from==T_SRC) ||
		      (medest && from==T_DEST) );
      // money
      if (money<0) {	// src gets money
	sprintf(srcmon, "$%d ", -money);
      }
      if (money>0) {	// dest gets money
	sprintf(destmon, "$%d ", money);
      }

      int nprops = 0;

      if (event==E_TRADE_OFFER) 
	sprintf(msg, "%s sent trade offer to %s",
		namefrom, nameto);
      else
	sprintf(msg, "%s changed trade offer in trade with %s",
		namefrom, nameto);

      appendGameMsg(msg);
      appendChatMsg(msg);

      sprintf(msg, "  %s offer%s %s%sthe following properties:",
	      namefrom, (mefrom)?"":"s", moneyfrom, (*moneyfrom)?"and ":"");
      appendGameMsg(msg);

      nprops=0;
      for (i=0; i<BOARD_N_FIELDS; ++i) {
	if (packet->prop[i] == from) {
	  sprintf(msg, "    %s", field_locale(i)->name);
	  appendGameMsg(msg);
	  nprops++;
	}
      }
      if (nprops==0) {
	appendGameMsg("    <no properties>");
      } 

      sprintf(msg, "  %s demand%s %s%sthe following properties:",
	      namefrom, (mefrom)?"":"s", moneyto, (*moneyto)?"and ":"");
      appendGameMsg(msg);
 
      nprops=0;
      for (i=0; i<BOARD_N_FIELDS; ++i) {
	if (packet->prop[i] == to) {
	  sprintf(msg, "    %s", field_locale(i)->name);
	  appendGameMsg(msg);
	  nprops++;
	}
      }
      if (nprops==0) {
	appendGameMsg("    <no properties>");
      } 

    }
    return;

  case E_TRADE_AGREE:
    // doesn't occur (so far)
/*      sprintf(msg, "trade between %s and %s: %s agreed",  */
//      assert(0);
    return;

  case E_TRADE_REJECT:
    {
      // for reject, offer_from doesn't contain the actor
      // but the offer which was rejected !!
      int to = packet->offer_from;
      int from = (to==T_SRC)?T_DEST:T_SRC;	// the other
      const char *namefrom = (from==T_SRC)?namesrc:namedest;
      const char *nameto = (from==T_SRC)?namedest:namesrc;
      sprintf(msg, "%s rejected trade with %s",
	      namefrom, nameto);
      appendGameMsg(msg);
      appendChatMsg(msg);
    }
    return;

  case E_NOEVENT:
    // only update trades, do nothing here
//      sprintf(msg, "trade info received: noevent");
    return;

  default:
    assert(0);		// bug
  }

}

void KapView::handle_field_info(struct packet_field_info *packet)
{
#define MSG_POS	(msg+strlen(msg))
#define DET_POS (details+strlen(details))

  // location of field
  int loc = packet->location;
  const char *locfieldname = field_locale(loc)->name;

  // action
  int event = packet->event;

  // status (all infos from packet)
  int stat = packet->statusflags;
  int act = packet->actionflags;
  int ownerid = (stat & IS_OWNED) ? packet->owner : NOBODY;
  int nh = packet->nhouses;
  int maxh = STREET_MAX_HOUSES;

  char msg[MSG_SIZE];
  char head[MSG_SIZE];
  char details[MSG_SIZE];
  *msg = '\0';
  *head = '\0';
  *details = '\0';

  // for actions with properties (get owner (as player) and ownername)
  const Property *prop = 
    dynamic_cast<const Property *>(getGame()->board()->field(loc));
  const Player *owner = (ownerid!=NOBODY)? getGame()->player(ownerid) : 0;
  bool mine = (ownerid == getGame()->myPID());
//    const char *ownername = (mine)? "you" : (owner)? owner->name() :"nObOdY";
  const char *ucownername = (mine)? "You" : (owner)? owner->name() :"NOBODY";

  // for these actions, plr info must be received before field info
  if (event == E_BUY_PROP ||
      event == E_MORTGAGE_PROP || event == E_UNMORTGAGE_PROP ||
      event == E_BUY_HOUSES || event==E_SELL_HOUSES) {
    assert(prop);
    // this is a hack to ensure that plr_info is received first!!
    assert(plr_info_received);	
    plr_info_received = false;
  }    

  // for actions with houses (get number of houses bought etc.)
  const Street *street = 
    dynamic_cast<const Street *>(getGame()->board()->field(loc));

  // first update gui (TODO: at end of function, but there are returns!)
  _board->field(loc)->update_status();

  // this is perhaps a little overhead (because it was already done in 
  // handle_player_info in some cases), but we want to be sure, that 
  // everything is really updated 
  _myinfo->update();
  _myhold->update();
  _plrinfo->update();

  switch (event) {

  case E_NOEVENT:
    // only update field, do nothing here
//      sprintf(msg, "field info received: noevent");
//      break;
    return;
  case E_ENTER_FIELD:

    if (stat) {	// some status flags are set
      sprintf(msg, "This field ");
      
      // owner etc.
      if (stat & IS_OWNED) {
	if (stat & IS_YOURS)
	  sprintf(MSG_POS, "is yours");
	else 
	  sprintf(MSG_POS, "is owned by %s", owner->name());

	sprintf(details, msg);

	// mortgaged ?
	if (stat & IS_MORTGAGED) {
	  sprintf(MSG_POS, " and is mortgaged");
	  sprintf(DET_POS, " and is mortgaged.");
	  if (!(stat & IS_YOURS))
	    sprintf(details+strlen(details), "\nYou are lucky, no rent.");
	}

	// monopoly (only for text window)
	if (stat & IS_MONOP)
	  sprintf(MSG_POS, " and is monopoly (%d %s)",
		  (nh<=maxh)?nh:1, (nh<=maxh)?"houses":"hotel");
      }
      sprintf(MSG_POS, ".");
      appendGameMsg(msg);

      _gameinfo->append_details(details);
    }

    if (act & CAN_BUY) { // some of the specified action flags are set?
      sprintf(msg, "You can ");
      
      // can buy ?
      if (act & CAN_BUY) {
	assert(prop);
//  	sprintf(MSG_POS, "buy it for $%d. (budget $%d)", prop->price(),
//  		getGame()->myPlr()->budget());
	sprintf( MSG_POS, "buy it for $%d.", prop->price() );
	getApp()->enable_actions(A_PROPERTY_BUY);  // repeated here for safety
      }

      sprintf(details, msg);

      appendGameMsg(msg);
      _gameinfo->append_details(details);

    }      
    //      _gameinfo->set_buttons(BUT_BUY);
    return;

  case E_BUY_PROP:
    sprintf(msg, "%s bought %s (%d) for $%d", 
	    ucownername, locfieldname, loc, -owner->last_transaction());
//      if (mine) sprintf(msg+strlen(msg), "\n\t-->(your new budget: $%d)", 
//  		      prop->owner()->budget());
    appendGameMsg(msg);

    sprintf(head, "%s bought %s (%d).", ucownername, locfieldname, loc);
    sprintf(details, "%s paid $%d for it.", ucownername, 
	    -owner->last_transaction());
    
    _gameinfo->message(MSG_INFO, head, details);
    return;
    
  case E_MORTGAGE_PROP:
    sprintf(msg, "%s mortgaged %s (%d) and got $%d for it",
	    ucownername, locfieldname, loc, owner->last_transaction());
//      if (mine) sprintf(msg+strlen(msg), "\n\t-->(your new budget: $%d)", 
//  		      prop->owner()->budget());
    appendGameMsg(msg);

    sprintf(head, "%s mortgaged %s (%d).", ucownername, locfieldname, loc);
    sprintf(details, "%s got $%d for it.", ucownername, 
	    owner->last_transaction());
    _gameinfo->message(MSG_INFO, head, details);
    return;

  case E_UNMORTGAGE_PROP:
    sprintf(msg, ("%s unmortgaged %s (%d) and paid $%d "
		  "(10%% interest included)"),
	    ucownername, locfieldname, loc, -owner->last_transaction());
//      if (mine) sprintf(msg+strlen(msg), "\n\t-->(your new budget: $%d)", 
//  		      prop->owner()->budget());
    appendGameMsg(msg);

    sprintf(head, "%s unmortgaged %s (%d).", ucownername, locfieldname, loc);
    sprintf(details, "%s paid $%d for it (10%% interest included).", 
	    ucownername, -owner->last_transaction());
    _gameinfo->message(MSG_INFO, head, details);
    return;

  case E_BUY_HOUSES:
    {
      assert(street);
      
      int addhouses = -owner->last_transaction() / street->house_price();
      
      sprintf(msg, "%s bought ", ucownername);
      
      // what was bought ?
      if (addhouses<=STREET_MAX_HOUSES)	// houses bought
	sprintf(msg+strlen(msg), "%d house%s ", 
		addhouses, (addhouses!=1)?"s":" ");
      else	// hotel bought
	sprintf(msg+strlen(msg), "1 hotel  ");
      
      sprintf(msg+strlen(msg), "on %s (%d) for $%d.", 
	      locfieldname, loc, -owner->last_transaction());
//        if (mine) sprintf(msg+strlen(msg), "\n\t-->(your new budget: $%d)", 
//  			street->owner()->budget());
      
      appendGameMsg(msg);
      
      // in gameInfo
      // use msg as buffer (!)
      sprintf(msg, "%s (%d):", locfieldname, loc);
      sprintf(details, "%-20.20s", msg);
      // what was bought ?
      if (addhouses<=STREET_MAX_HOUSES)	// houses bought
	sprintf(details+strlen(details), "%d house%s ", 
		addhouses, (addhouses!=1)?"s":"");
      else	// hotel bought
	sprintf(details+strlen(details), "1 hotel ");
      sprintf(details+strlen(details), "for $%d.", 
	      -owner->last_transaction());
      
      // set new details text in gameinfo, if this is the first line that deals
      // with houses buy event
      if (!append_msg_houses_event) {
	_gameinfo->set_details(MSG_INFO, details);
	append_msg_houses_event = true;
      } else
	_gameinfo->append_details(details);

      // we know, that there will be a monop_info packet after all streets
      // that will set the headline !! Make sure !!

      return;

    }
  case E_SELL_HOUSES:
    {
      assert(street);

      // if houses are sold, owner gets only half the house price -> factor 2
      int subhouses = 2 * owner->last_transaction() / street->house_price();
//        assert(subhouses<=street->nhouses());

      // initialize details with fieldname (msg used as buffer!)
      sprintf(msg, "%s (%d):", locfieldname, loc);
      sprintf(details, "%-25.25s", msg);

      // initialize msg
      sprintf(msg, "%s sold ", ucownername);

      // what was sold ?
      if (subhouses<=STREET_MAX_HOUSES) {	// houses sold
	sprintf(msg+strlen(msg), "%d house%s ", 
		subhouses, (subhouses!=1)?"s":" ");
	sprintf(details+strlen(details), "%d house%s ", 
		subhouses, (subhouses!=1)?"s":" ");
      } else {	// hotel sold
	sprintf(msg+strlen(msg), "1 hotel  ");
	sprintf(details+strlen(details), "1 hotel  ");
      }
      sprintf(msg+strlen(msg), "on %s (%d) for $%d.", 
	      locfieldname, loc, owner->last_transaction());
      sprintf(details+strlen(details), "for $%d.",
	      owner->last_transaction());
//        if (mine) sprintf(msg+strlen(msg), "\n\t-->(your new budget: $%d)", 
//  			street->owner()->budget());

      appendGameMsg(msg);
      
      // set new details text in gameinfo, if this is the first line that deals
      // with houses event
      if (!append_msg_houses_event) {
	_gameinfo->set_details(MSG_INFO, details);
	append_msg_houses_event = true;
      } else
	_gameinfo->append_details(details);

      // we know, that there will be a monop_info packet after all streets
      // that will set the headline !! Make sure !!
      
      return;
    }
  
  case E_NEW_MONOP:
  case E_MONOP_CHANGED:
  case E_MONOP_DESTROYED:
    // handled with player info
/*      sprintf(msg, "field info received: monop event"); */
    return;

  case E_OWNER_CHANGED:
    // handled with trade packet
    return;
  default:
//      sprintf(msg, "field info received: unknown event: %d", event);
    break;

  }

  sprintf(details, msg);

  appendGameMsg(msg);
  _gameinfo->append_details(details);
    
#undef MSG_POS
#undef DET_POS

}

void KapView::handle_monop_info(struct packet_monop_info *packet)
{
#define MSG_POS	(msg+strlen(msg))
  
  // who is owner ?
  const Player *owner =  
    (getGame()->nobody(packet->owner)) ? 0 : getGame()->player(packet->owner);
  bool me = (owner == getGame()->myPlr());
  const char *name = (me)? "You" : (owner)? owner->name(): 0;

  // which monop ?
  int mid = packet->monopid;
  const char *locmonopname = monop_locale(mid).color;

  // action
  int event = packet->event;

  // TODO: emit signal monopChanged()
//    getApp()->updateHousesDialog(packet);

  char msg[MSG_SIZE];
  char details[MSG_SIZE];
  char head[MSG_SIZE];

  switch (event) {

  case E_NOEVENT:
    // only update monop, do nothing here
//      sprintf(msg, "monop info received: noevent");
//      break;
    return;
  case E_NEW_MONOP:
    assert(owner);
    {    
      sprintf(msg, "A new monopoly has evolved:\n"
	      "%s aquired all streets in %s.",
	      name, locmonopname);
      sprintf(details, "New monopoly!\n"
	      "%s aquired all streets in %s.",
	      name, locmonopname);
    }
    break;
    
  case E_MONOP_CHANGED:
    assert(owner);
    {
      sprintf(msg, "The %s monopoly has changed owner:\n"
	      " %s aquired it.", locmonopname, name);
      sprintf(details, "Monopoly %s changed owner: \n"
	      "%s aquired it.", locmonopname, name);
    }
    break;

  case E_MONOP_DESTROYED:
    assert(owner == 0);
    sprintf(msg, "The %s monopoly has been destroyed.",
	    locmonopname);
    sprintf(details, msg);
    break;

  case E_BUY_HOUSES:
  case E_SELL_HOUSES:
    assert(owner);
    sprintf(head, "%s %s buildings on %s.", name,
	    (event==E_BUY_HOUSES)? "bought":"sold", locmonopname);
    // this is necessary, because detail info was already set in 
    // handle_field_info
    _gameinfo->set_headline(head);
    append_msg_houses_event = false;
    return;

  default:
//      sprintf(msg, "monop info received: unknown event");
    break;

  }

  appendGameMsg(msg);
  _gameinfo->append_details(details);

#undef MSG_POS

}

void KapView::handle_game_init_info(struct packet_game_init_info */*packet*/)
{
//    int i;

  // _dice->set_dice(getGame()->dice1(), getGame()->dice2());
  // etc.
}

void KapView::handle_game_info(struct packet_game_info */*packet*/)
{
  int i;

  // initialize board
  _board->init_board(getGame()->board());

  // initialize players and set them to correct location
  for (i=0; i<getGame()->nplayers(); ++i) {
    _plr[i] = new PlayerView(this, _board, getGame()->player(i));
    _plr[i]->update_status(E_NOEVENT);
  }

//    myGplr = _plr[getGame()->myPID()];
  // note: I wanted to put the creation of the players in handle_game_init_info
  // but then it is before call of init_board

//    _myinfo->set_tokenColor(_plr[getGame()->myPID()]->tokenColor());
  _myinfo->init();
  _myinfo->update();

  _myhold->init();
  _myhold->update();

  _plrinfo->init();
  _plrinfo->update();

  // update fields
  for (i=0; i<BOARD_N_FIELDS; ++i)
    _board->field(i)->update_status();

}

void KapView::handle_player_info(struct packet_player_info *packet)
{
  // who is acting ?
  int plrid = packet->playerid;
  bool me = (plrid == getGame()->myPID());
  const char *name = (me)? "you" : getGame()->player(plrid)->name();
  const char *ucname = (me)? "You" : name;

  // where ?
  int loc = packet->location;
  const char *locfieldname = field_locale(loc)->name;

  // action
  int event = packet->event;
  int transaction = packet->transaction;

  char msg[MSG_SIZE];
  char head[MSG_SIZE];
  char details[MSG_SIZE];
  *msg = '\0';
  *head = '\0';
  *details = '\0';

  // for actions with properties
  const Property *prop =  
    (loc != JAIL) 
    ? dynamic_cast<const Property *>(getGame()->board()->field(loc))
    : 0;
  
  // TODO: move to the end of function?(currently, there are returns in case!)
  _plr[plrid]->update_status(event);
  _myinfo->update();
  _myhold->update();
  _plrinfo->update();
  
  switch (event) {

  case E_NOEVENT:
    // only update player, do nothing here
//      sprintf(msg, "player info received: noevent");
//      break;
    return;

  case E_YOUR_TURN:
    // already handled in handle_your_turn
    return;

  case E_MOVE:
  case E_MOVE_BACKWARDS:
    sprintf(msg, "That puts %s on %s (%d).", name, locfieldname, loc);
    appendGameMsg(msg);
    // this is the first message after dicing
    _gameinfo->set_details(MSG_INFO, msg);
    return;

  case E_MOVE_PASS_GO:
    sprintf(msg, "%s passed %s and collected $%d.\n"
	    "That puts %s on %s (%d).",
	    ucname, field_locale(GO)->name, transaction,
	    name, locfieldname, loc);
//  field_locale(GO).name
    appendGameMsg(msg);
    // this is the first message after dicing
    _gameinfo->set_details(MSG_INFO, msg);
    return;
    
  case E_BUY_PROP:
  case E_MORTGAGE_PROP:
  case E_UNMORTGAGE_PROP:
  case E_BUY_HOUSES:
  case E_SELL_HOUSES:
    plr_info_received = true;
    // handled when field info is received
    return;

  case E_PAY_RENT:
  case E_SHOULD_PAY_RENT:
    {
      assert(prop);
      const char *predicate = (event == E_PAY_RENT)? "paid" : "should pay";
      if (getGame()->myPlr() == prop->owner()) {	// I got the rent

	sprintf(msg, "%s %s you a rent of $%d.",
		ucname, predicate, -transaction);

      } else {	// I paid the rent or I'm not concerned
	
	sprintf(msg, "%s %s a rent of $%d to %s.",
		ucname, predicate, -transaction, prop->owner()->name());
      }
    }
    break;

  case E_RECV_RENT:
    // everything is already handled in E_PAY_RENT
    return;

  case E_PAY_TAX:
    sprintf(msg, "%s paid a tax of $%d.", ucname, -transaction);
    break;
  case E_SHOULD_PAY_TAX:
    sprintf(msg, "%s should pay a tax of $%d.", ucname, -transaction);
    break;
    
  case E_CARD_COLLECT:
    sprintf(msg, "%s collected $%d from the bank." , 
	    ucname, transaction);
    break;
  case E_CARD_COLLECT_FROM_ALL:
    sprintf(msg, "%s collected $%d from each player, total $%d." , 
	    ucname, transaction, (getGame()->nplayers()-1)*transaction );
    break;
  case E_CARD_PAY:
    sprintf(msg, "%s paid $%d to the bank." , 
	    ucname, -transaction);
    break;
  case E_CARD_SHOULD_PAY:
    sprintf(msg, "%s should pay $%d to the bank." , 
	    ucname, -transaction);
    break;

  case E_CARD_PAY_TO_ALL:
    sprintf(msg, "%s paid $%d to each player, total $%d." , 
	    ucname, -transaction/(getGame()->nplayers()-1), -transaction );
    break;

  case E_CARD_JAIL_FREE:
    sprintf(msg, "%s got a jail free card.", ucname);
    break;
  case E_CARD_GOTO_JAIL:
    sprintf(msg, "%s has to go to jail.", ucname);
    break;
    
  case E_CARD_RENOVATE_BUILDINGS:
    sprintf(msg, "%s paid a total of $%d for all of %s buildings.",
	    ucname, -transaction, (me)?"your":"his/her");
    break;
  case E_CARD_SHOULD_RENOVATE_BUILDINGS:
    sprintf(msg, "%s should pay a total of $%d for all of %s buildings.",
	    ucname, -transaction, (me)?"your":"his/her");
    break;
    
  case E_GOTO_JAIL_ENTERED_FIELD:
    sprintf(msg, "%s %s to go in jail.", ucname, (me)?"have":"has");
    break;
  case E_GOTO_JAIL_ROLLED_DOUBLES:
    sprintf(msg, "%s rolled 3 doubles, that puts %s in jail.",
	    ucname, (me)?"you":"him/her");
    break;

  case E_LEAVE_JAIL_ROLLED_DOUBLES:
    sprintf(msg, "%s rolled doubles and left jail (no fine).", ucname);
    break;
  case E_LEAVE_JAIL_NORMAL:
    sprintf(msg, "%s left jail after 3 turns and paid a fine of $%d.", 
	    ucname, -transaction);
    break;
  case E_LEAVE_JAIL_USE_CARD:
    sprintf(msg, "%s used jail free card and left jail.", ucname);
    break;
    
  case E_LEAVE_JAIL_PAY_FINE:
    sprintf(msg, "%s paid a fine of $%d and left jail.", 
	    ucname, -transaction);
    break;

  case E_STAY_IN_JAIL:
    sprintf(msg, "%s stay%s in jail for another turn.", ucname, (me)?"":"s");
    break;

  case E_TRADE_FINISHED:
    // handled with trade packet
    return;

  case E_GO_BANKRUPT:
    sprintf(msg, "%s DECLARE%s %sSELF BANKRUPT", ucname, 
	    (me)?"":"S", (me)?"YOUR":"HIM");
    sprintf(head, msg);
    *details = '\0';
    appendGameMsg(msg);
    appendChatMsg(msg);
    _gameinfo->message(MSG_INFO, head, details);
    // TODO: after this there is a next_turn packet immediately, so these
    // info messages about the bankruptcy can't be read. Fix this.
    // this will be resolved when dialog windows are used
    return;

  case E_BANKRUPT_CREDITOR:
    sprintf(msg, "all cash and holdings go to %s", (me)?"you":ucname);
    break;

  case E_SETTLE_DEBTS:
    sprintf(msg, "%s settled %s debts.", ucname, (me)?"your":"his/her");
    break;

  case E_DIE:
    sprintf(msg, "%s died.", ucname);
    break;

  case E_LOST_CONNECTION:
//      sprintf(msg, "*** lost connection to %s ***", name);
//      break;
    // message comes already from server
    return;

  default:
//      sprintf(msg, "player info received: unknown event: %d", event);
    return;
  }

  sprintf(details, msg);

  appendGameMsg(msg);

  _gameinfo->append_details(details);

}


struct monop_locale_struct KapView::monop_locale(int monopID)
{
  assert(monopID>=0 && monopID<BOARD_N_MONOPS);
  return ::monop_locale[monopID];
}

const ThemeFieldEntry *KapView::field_locale(int fieldID) const
{
  return _theme->field(fieldID);
}

const ThemeMoneyEntry *KapView::money_locale() const
{
  return _theme->money();
}

QString KapView::money_locale(const int money) const
{
  QString moneystr;
  moneystr.setNum( (int) round( money_locale()->factor * money ) );
  moneystr.prepend( money_locale()->prefix );
  return moneystr;
}

const ThemeDiceAnimEntry *KapView::diceAnim() const
{
  return _theme->diceAnim();
}

//  void KapView::locales_eliminate_hyphens()
//  {
//    int id;
//    char *src, *dest;

//    for (id=0; id<LAST_LOC; ++id) {
//      src = ::field_locale_raw[id].name;
//      dest = ::field_locale[id].name;
//      for (; *src; src++, dest++) {
//        *dest = *src;
//        if (*src == '_') *dest = ' ';
//        if (*src == '-') { src++; *dest = *src; }
//      }
//      *dest = '\0';
//    }
  
//  }



struct card_locale_struct KapView::cchest_card_locale(int cchestID)
{
  assert(cchestID>=0 && cchestID<GAME_N_COMCHESTS);
  return ::cchest_card_locale[cchestID];
}

struct card_locale_struct KapView::chance_card_locale(int chanceID)
{
  assert(chanceID>=0 && chanceID<GAME_N_CHANCES);
  return ::chance_card_locale[chanceID];
}
 
//
// slots
//

void KapView::slotIdleCycle()
{
  // takes half a sec, one blink of current player
//    static int ncalls = 0;
  PlayerView *curplr = currPlr();
  bool me = ( (curplr == myPlrView()) ? true : false );

  // break the loop, if desired (by call of slotFinishIdleCycle before)
  if (_finishIdleCycle) {
    if (curplr) {
      curplr->show();  // in case tokens are currently hidden
      if (me) _myinfo->showToken();
      else _plrinfo->showToken(getGame()->currPlr()->id());
    }
    _finishIdleCycle = false; // reset for the next time
    emit idleCycleFinished();
    return;
  }

  if (!curplr) {  // we have no current turn yet
    _idleTimer.start(500, TRUE);
    return;
  }

  assert(getGame()->currPlr());
  
  if (curplr->isHidden()) {  // show tokens and wait for 300 msecs
    curplr->show();
    if (me) _myinfo->showToken();
    else _plrinfo->showToken(getGame()->currPlr()->id());
    _idleTimer.start(600, TRUE);
  } else {  // hide tokens and wait for 200 msecs
    curplr->hide();
    if (me) _myinfo->hideToken();
    else _plrinfo->hideToken(getGame()->currPlr()->id());
    _idleTimer.start(400, TRUE);
  }
}

void KapView::slotFinishIdleCycle()
{
  _finishIdleCycle = true;
}




