// -*- C++ -*-

/***************************************************************************
                          gameinfo.cpp  -  description
                             -------------------
    begin                : Mon Jun 11 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//
// System
//

#include <assert.h>

using namespace std;

//
// Qt
//

#include <qcanvas.h>
#include <qpoint.h>
#include <qpainter.h>

//
// Game
//

#include "field.h"

//
// View
//

#include "kapview.h"
#include "gameinfo.h"
#include "dieview.h"


/**
 * we need a rounded rectangle
 */
class QCanvasRoundRectangle : public QCanvasRectangle {

public:  

  QCanvasRoundRectangle(int x, int y, int width, int height, QCanvas* canvas)
    : QCanvasRectangle(x, y, width, height, canvas) {}

protected:

  void drawShape(QPainter &p) {
    // we want circles at the corners
    int xRnd = 10;
    // heuristics: height = 20 -> yRnd = 90, height = 100 -> yRnd = 20
    double slope = (20.0-90.0)/(100.0-20.0); 
    double intersect = (90.0 - slope * 20.0);
    int yRnd = (int) ( slope * height() + intersect );
    if (yRnd < 20) yRnd = 20;
    p.drawRoundRect((int)x(), (int)y(), width(), height(), xRnd, yRnd);
  }

};

/**
 * some fixed colors
 */
const QColor chanceColor = QColor(200,255,250);	//light blue
const QColor cchestColor = QColor(255,240,221); //baby face color



GameInfo::GameInfo(KapView *view, QCanvas *canvas, QWidget *parent)
  : QCanvasView(canvas, parent), _view(view), _canvas(canvas),
    _infoColor(_view->centerColor().light(200)),
    _die1_moving(false), _die2_moving(false)
{
  const int zShadow = 2;
  const int zBg = 3;
  const int zText = 4;
  const int zCardShadow = 7;
  const int zCardBg = 8;
  const int zCardText = 9;
  const int zDice = 10;

//    setAlignment(AlignLeft);
//    setTitle("Game Info");
    setFrameStyle( NoFrame );
    setLineWidth(0);
    setMargin(0);
//    setBackgroundColor(_view->centerColor().light(200));
//    _canvas->setBackgroundColor(_infoColor);

  // initialize some fonts needed for later use
  QFont headfont = font();
  QFont detailsfont = font();
  QFont cardfont = font();
  headfont.setBold(true);
  headfont.setPointSize(headfont.pointSize() + 2);
  QFontMetrics headfm(headfont);
  QFontMetrics detailsfm(detailsfont);
  QFontMetrics cardfm(cardfont);

  // create the headline elements (background + shadow + text)
  _headlineBg = new QCanvasRoundRectangle(0,0, _canvas->width()*4/5,
					  headfm.height() + 10, _canvas);
  _headlineBg->setZ(zBg);
  _headlineBg->setPen(black);
  _headlineBg->setBrush(QBrush(_infoColor));
  _headlineShadow = new QCanvasRoundRectangle(0,0, _headlineBg->width(),
					      _headlineBg->height(), _canvas);
  _headlineShadow->setZ(zShadow);
  _headlineShadow->setPen(NoPen);
  _headlineShadow->setBrush(QBrush(_view->centerColor().dark(150)));
  _headline = new QCanvasText("This is the headline", headfont, _canvas);
  _headline->setTextFlags(Qt::AlignCenter);
  _headline->setZ(zText);
  
  // create the details elements (background + shadow + text)
  _detailsBg = new QCanvasRoundRectangle(0,0, _canvas->width()*4/5, 
					 4 * detailsfm.height() + 10, 
					 _canvas);
  _detailsBg->setZ(zBg);
  _detailsBg->setPen(black);
  _detailsBg->setBrush(QBrush(_infoColor));
  _detailsShadow = new QCanvasRoundRectangle(0,0, _detailsBg->width(), 
					     _detailsBg->height(), _canvas);
  _detailsShadow->setZ(zShadow);
  _detailsShadow->setPen(NoPen);
  _detailsShadow->setBrush(QBrush(_view->centerColor().dark(150)));
  _details = new QCanvasText("This is a text spanning multiple rows\n"
			     "two\nthree\nfour", detailsfont, _canvas);
  _details->setTextFlags(Qt::AlignCenter);
  _details->setZ(zText);

  // create the dice
  _die1 = new DieView(_view, _canvas, this);
  _die1->move(30,100);
  _die1->setAnimated(false);
  _die1->setZ(zDice);

  _die2 = new DieView(_view, _canvas, this);
  _die2->move(30,150);
  _die2->setAnimated(false);
  _die2->setZ(zDice);

  // create the cards
  _cardBg = new QCanvasRoundRectangle(0,0, 10, 10, _canvas);
  _cardBg->setZ(zCardBg);
  _cardBg->setPen(red);
  _cardBg->setBrush(NoBrush);
  _cardShadow = new QCanvasRoundRectangle(0,0, 10, 10, _canvas);
  _cardShadow->setZ(zCardShadow);
  _cardShadow->setPen(NoPen);
  _cardShadow->setBrush(QBrush(_view->centerColor().dark(150)));
  _cardText = new QCanvasText("Card text that can\n"
			      "exceed more than\n"
			      "one line",
			      cardfont, _canvas);
  _cardText->setZ(zCardText);
  _cardText->setTextFlags(Qt::AlignCenter);
  _cardText->setColor(QColor(50,0,0));

  setResizePolicy(Manual);

  // show all elements
  _headlineBg->show();
  _headlineShadow->show();
  _headline->show();
  _detailsBg->show();
  _detailsShadow->show();
  _details->show();
  _die1->show();
  _die2->show();

  show();
}

GameInfo::~GameInfo()
{
}

void GameInfo::message(int msg_type, const char *headline, const char *details)
{
  set_headline(headline);
  set_details(msg_type, details);
//    set_buttons(button0, button1, button2);
  _canvas->update();
}


void GameInfo::set_headline(const char *header)
{
  _headline->setText(_view->replace_locales(header));
  _canvas->update();
}

void GameInfo::set_details(int /*msg_type*/, const char *details)
{
  _details->setText(_view->replace_locales(details));
  _canvas->update();
//    set_icon(msg_type);
}

void GameInfo::append_details(const char *details)
{
  const char *txt =  _details->text();
  bool with_newline = false;

  if (*txt)	// text wasn't empty -> append newline at end of existing text
    with_newline = true;

  QString newtxt = ( QString(txt) + QString( (with_newline)?"\n":"") + 
		     QString(details) );

  _details->setText(_view->replace_locales(newtxt));
  _canvas->update();
}


void GameInfo::set_dice(const int val1, const int val2)
{
  _view->getApp()->animationStarted();
  _die1_moving = true;
  _die2_moving = true;
  
  _canvas->setAdvancePeriod(10);
  _die1->move(_canvas->width() - 30, 30);
  _die1->setVelocity(-2,1);  // velocity recalculated in DieView::advance
  _die1->setAnimated(true);
  _die1->setValue(val1);

  _die2->move(_canvas->width() - 30, 70);
  _die2->setVelocity(-2,1);  // velocity recalculated in DieView::advance
  _die2->setAnimated(true);
  _die2->setValue(val2);
  // animation is stopped when both dice report end of movement
}

void GameInfo::die_stopped(const DieView *die)
{
  // ignore this call when both dice are already in rest
  if (!(_die1_moving || _die2_moving)) return;
  
  if (die == _die1) _die1_moving = false;
  else if (die == _die2) _die2_moving = false;

  // if both dice are in rest now -> stop animation in app
  if (!(_die1_moving || _die2_moving)) {
    _canvas->setAdvancePeriod(-1);
    _view->getApp()->animationFinished();
  }
  
}


//  CardView::CardView(KapView *view, QWidget *parent)
//      : QLabel(parent), _view(view)
//  {
//    setText("first line\nsecond line\nthird line");
//    setAlignment(AlignCenter);
//    setFrameStyle( Box | Sunken );
//    setFixedSize( QSize(250,120) );
//    hide();
//  }

//  CardView::~CardView()
//  {
//  }



void GameInfo::draw_card( const int type, const int card_no )
{
  switch (type) {
  case COMCHEST:
//      _cardBg->setBackgroundColor(cchestColor);
    _cardText->setText(_view->replace_locales(_view->cchest_card_locale(card_no).msg));
    _cardBg->setBrush(QBrush(cchestColor));
    _cardBg->setPen(red);
    break;
  case CHANCE:
//      _cardBg->setBackgroundColor(chanceColor);
    _cardText->setText(_view->replace_locales(_view->chance_card_locale(card_no).msg));
    _cardBg->setBrush(QBrush(chanceColor));
    _cardBg->setPen(blue);
    break;
  default:
    assert(0);
  }
  _cardShadow->show();
  _cardBg->show();
  _cardText->show();

  _canvas->update();
}

void GameInfo::hide_card()
{
  _cardText->hide();
  _cardBg->hide();
  _cardShadow->hide();

  _canvas->update();
}

/*
 * protected methods
 */

void GameInfo::resizeEvent( QResizeEvent * )
{
  QSize vpSize = viewportSize(size().width(), size().height());
  _detailsBg->hide();
  canvas()->resize(vpSize.width(), vpSize.height());
  _detailsBg->show();

  QFontMetrics headfm(_headline->font());
  QFontMetrics detailsfm(_details->font());
  QFontMetrics cardfm(_cardText->font());

  // some sizes
  int canvasW = _canvas->width();
  int canvasH = _canvas->height();
  int headlineW = canvasW * 4/5;
  int headlineH = headfm.height() + 10;
  int detailsW = canvasW * 4/5;
//    int detailsH = 4 * detailsfm.height() + 10;
  int detailsH = canvasH/2;
  int cardW = 20 * cardfm.width("M") + 10;
  int cardH = 4 * cardfm.height() + 10;

  // some positions (top left corner)
  int headlineX = canvasW/2 - headlineW/2;
  int headlineY = 20;
  int detailsX = canvasW/2 + 20 - detailsW/2;
  int detailsY = headlineY+headlineH + 20;
  int cardX = canvasW/2 + 20 - cardW/2;
  int cardY = canvasH - 20 - cardH;

  // adjust size of elements
  _headlineBg->setSize(headlineW, headlineH);
  _headlineShadow->setSize(headlineW, headlineH);
  _detailsBg->setSize(detailsW, detailsH);
  _detailsShadow->setSize(detailsW, detailsH);
  _cardBg->setSize(cardW, cardH);
  _cardShadow->setSize(cardW, cardH);

  // move elements to the right place
  _headlineBg->move(headlineX, headlineY);
  _headlineShadow->move(headlineX+3, headlineY-3);
  _detailsBg->move(detailsX, detailsY);
  _detailsShadow->move(detailsX+3, detailsY-3);
  _cardBg->move(cardX, cardY);
  _cardShadow->move(cardX+5, cardY-5);

  _headline->move(headlineX+headlineW/2, headlineY+headlineH/2);
  _details->move(detailsX+detailsW/2, detailsY+detailsH/2);
  _cardText->move(cardX+cardW/2, cardY+cardH/2);

  _canvas->update();
  update();
}

void GameInfo::set_icon(int /*msg_type*/)
{
//    switch (msg_type) {
//    case MSG_INFO:
//      _iconLab->setPixmap(_infoPixmap);
//      break;
//    case MSG_COMMAND:
//      _iconLab->setPixmap(_cmdPixmap);
//      break;
//    case MSG_QUERY:
//      _iconLab->setPixmap(_queryPixmap);
//      break;

//    default:
//      assert(0);	// undesired -> bug	(TODO: default INFO)
//    }
}
