// -*- C++ -*-

/***************************************************************************
                          inputline.cpp  -  description
                             -------------------
    begin                : Mon Jun 11 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//
// System
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <cmath>
//
// Qt
//

#include <qlineedit.h>
#include <qlayout.h>
#include <qframe.h>
#include <qkeycode.h>

//
// App
//

#include "kapitalist.h"
#include "global.h"
#include "events.h"
#include "log.h"
#include "mem.h"
#include "console.h"

//
// Game
//

#include "kapgame.h"
#include "board.h"
#include "field.h"
#include "property.h"
#include "street.h"
#include "railroad.h"
#include "utility.h"
#include "monop.h"
#include "trade.h"

//
// View
//

#include "kapview.h"
#include "themedata.h"

#include "inputline.h"


#define MAX_CMD_LEN MAX_PACKET_SIZE

const char *InputLine::horiz_line 
 = "-----------------------------------------";

const InputLine::command InputLine::commands[] = {
  {"help"},
  {"list"},
  {"quit"},
  {"chat"},
    
  {"info"},

  {"roll"},
  {"done"},
  {"property"},
  {"houses"},
  {"jail"},
  {"trade"},
    
  {"pay"},
  {"bankruptcy"},
  {"connect"},
  {"disconnect"},
    
  {"log"},
  {"crash"}
};

const InputLine::command InputLine::info_cmd[] = {
  {"board"},
  {"field"},
  {"player"}
};
  
  
const InputLine::command InputLine::prop_cmd[] = {
  {"buy"},
  {"mortgage"},
  {"unmortgage"}
};
  
  
const  InputLine::command InputLine::houses_cmd[] = {
  {"buy"},
  {"sell"}
};
  
const InputLine::command InputLine::jail_cmd[] = {
  {"pay"},
  {"card"}
};
  
const InputLine::command InputLine::trade_cmd[] = {
  {"offer"},
  {"reply"},
  {"agree"},
  {"reject"}
};


InputLine::InputLine(KapView *view, QWidget *parent, int mode)
  : QLineEdit(parent), _view(view),
    _frm(0), _mode(mode)
{
  if (_mode < 0 || _mode >= LAST_MODE) _mode = CHAT_MODE;

  _top = new QHBoxLayout(parent);
  setMinimumSize( sizeHint() );
  setFixedHeight( sizeHint().height() );
  connect(this, SIGNAL(returnPressed()), 
	  this, SLOT(handle_line_input()));

  _top->addWidget(this);
  
  _top->activate();

  show();
}

InputLine::~InputLine()
{
}

void InputLine::set_frame(UpDownFrame *frm)
{
  _frm = frm;
}

void InputLine::handle_line_input()
{
  QString line = text();
  
  assert(_frm);
  _frm->append_history(text());

  handle_stdin_input( const_cast<char *>( (const char *)(text()) ) );

  // remove text from visible line
  clear();
}


enum InputLine::command_id 
InputLine::command_named(char *token, int accept_ambiguity)
{
  int i;
  enum command_id found = CMD_UNRECOGNIZED;
  unsigned int len = strlen(token);

  for (i = 0; i < CMD_NUM; ++i) {
    if (strncmp(commands[i].name, token, len) == 0) {
      if (accept_ambiguity) {
	return (enum command_id) i;
      } else if (found != CMD_UNRECOGNIZED) {
        return CMD_AMBIGUOUS;
      } else if (strlen(commands[i].name) == len) {
	return (enum command_id) i;
      } else {
        found = (enum command_id) i;
      }
    }
  }
  return found;
}

int InputLine::lookup_subcommand(char *find, const struct command cmd[], 
				int cmd_last)
{
  int i, lastmatch = -1;
    
  for (i=0; i<cmd_last; i++) 
    if (!strncmp(find, cmd[i].name, strlen(find))) 
      {
	/* If name is a perfect match, not partial, assume
	 * that this is the command the user wanted. */
	if (!strcmp(find, cmd[i].name))
	  return i;
	if (lastmatch != -1)
	  return -2;
	lastmatch = i;
      }
    
  return lastmatch;
}


void InputLine::cmd_reply(enum command_id /*cmd*/,
			  int /*console_id*/, 
			  const char *format, ...)
{
  char line[MAX_CMD_LEN];

  va_list ap;
//    const char *cmdname = cmd < CMD_NUM ? commands[cmd].name :
//      cmd == CMD_AMBIGUOUS ? "(ambiguous)" :
//      cmd == CMD_UNRECOGNIZED ? "(unknown)" :
//      "(?!?)";  /* this case is a bug! */

  va_start(ap,format);
  /* (void)vsnprintf(line,MAX_CMD_LEN-1,format,ap); */
  /*
   * no snprintf() in ANSI C and I don't know how to do it otherwise;
   * then again, there are plenty of other places in which this overflow
   * exists so I won't solve it here until we have myvsnprintf() - rp
   */
  vsprintf(line,format,ap);
  va_end(ap);

  _view->appendCurrOut(line);
}

// TODO: this is only temporary !!!

void InputLine::crash_and_burn()
{
  cmd_reply(CMD_CRASH, C_GENFAIL, "Crashing and burning.");
  /* Who is General Failure and why is he crashing and
     burning my computer? :) -- Per */
  assert(0);
}

void InputLine::handle_stdin_input(char *str)
{
  char cmdname[MAX_CMD_LEN+1], arg[MAX_CMD_LEN+1], *cptr_s, *cptr_d;
  int i;
  enum command_id cmd;

  enum client_states client_state = _view->getApp()->client_state();

  /* Is it a comment or a blank line? */
  /* line is comment if the first non-whitespace character is '#': */
  for(cptr_s=str; *cptr_s && isspace(*cptr_s); cptr_s++);
  if(*cptr_s == 0 || *cptr_s == '#') {
    return;
  }

  // distinguish 3 types of input:
  // input with server/client prefix is treated independently of the mode
  // input with no prefix is treated as either chat or client command
  // depending on the mode

  // server commands:
  if (*cptr_s == SERVER_COMMAND_PREFIX) {
    // error when disconnected
    if (client_state < CLIENT_PRE_GAME_STATE) {
      cmd_reply(CMD_CHAT, C_FAIL, ("Please connect to server first") );
      return;
    }

    struct packet_generic_message pack;

    strncpy(pack.message, cptr_s, MSG_SIZE);
    pack.message[MSG_SIZE-1]='\0';
    send_packet_generic_message(_view->getApp()->conn(),
				PACKET_CHAT_MSG, &pack);
    return;
  }

  // treat input with no prefix depending on mode
  if (*cptr_s != CLIENT_COMMAND_PREFIX && *cptr_s != SERVER_COMMAND_PREFIX) {
    switch (_mode) {
    case CHAT_MODE:
      {  // send input as chat message to server
	// error when disconnected
	if (client_state < CLIENT_PRE_GAME_STATE) {
	  cmd_reply(CMD_CHAT, C_FAIL, ("Please connect to server first") );
	  return;
	}
	
	char tmpmsg[MSG_SIZE], newmsg[MSG_SIZE];
	chat_process_field_names(cptr_s, tmpmsg);
	chat_process_money(tmpmsg, newmsg);

	struct packet_generic_message pack;
	
	strncpy(pack.message, newmsg, MSG_SIZE);
	pack.message[MSG_SIZE-1]='\0';
	send_packet_generic_message(_view->getApp()->conn(),
				    PACKET_CHAT_MSG, &pack);
	return;
      }
    case GAME_MODE:
      // go on and interpret input as client command
      break;
    default: 
      assert(0);  // any other modes possible?
    }
  }    

  // client commands:
  if (*cptr_s == CLIENT_COMMAND_PREFIX) {
    cptr_s++;  // simply ignore this prefix and continue
  }

  for(cptr_s=str; *cptr_s && !isalnum(*cptr_s); cptr_s++);

  for(cptr_d=cmdname; *cptr_s && isalnum(*cptr_s) &&
	cptr_d < cmdname+sizeof(cmdname)-1; cptr_s++, cptr_d++)
    *cptr_d=*cptr_s;
  *cptr_d='\0';

  cmd = command_named(cmdname,0);
  if (cmd == CMD_AMBIGUOUS) {
    cmd = command_named(cmdname,1);
    cmd_reply(cmd,  C_SYNTAX,
	      "Warning: '%s' interpreted as '%s', but it is ambiguous. "
	      "Try '.help'.",
	      cmdname, commands[cmd].name);
  } else if (cmd == CMD_UNRECOGNIZED) {
    cmd_reply(cmd,  C_SYNTAX,
	      "Unknown command.  Try '.help'.");
    return;
  }

  for(; *cptr_s && isspace(*cptr_s); cptr_s++);
  strncpy(arg, cptr_s, MAX_CMD_LEN);
  arg[MAX_CMD_LEN]='\0';

  i=strlen(arg)-1;
  while(i>0 && isspace(arg[i]))
    arg[i--]='\0';

//    // disable connect
//    if (cmd==CMD_CONNECT)
//        cmd_reply(cmd, C_FAIL, ("You are already connected") );
    
  // disable some commands, if there is no connection and/or no game yet
  if ( client_state < CLIENT_GAME_RUNNING_STATE &&
       (cmd==CMD_INFO || cmd==CMD_ROLL || cmd==CMD_DONE || 
        cmd==CMD_PROPERTY || cmd==CMD_HOUSES || cmd== CMD_JAIL ||
	cmd==CMD_TRADE || cmd==CMD_PAY || cmd==CMD_BANKRUPTCY) ) {
    if (client_state == CLIENT_BOOT_STATE)
      cmd_reply(cmd, C_FAIL, ("Please connect to server first") );
    else
      cmd_reply(cmd, C_FAIL, ("Please wait until game starts") );
    return;
  }

  switch(cmd) {
  case CMD_HELP:
    show_help();
    break;
  case CMD_LIST:
    show_players();
    break;
  case CMD_QUIT:
    dlg_quit();
    break;
  case CMD_CHAT:
    chat(arg);
    break;

  case CMD_INFO:
    info(arg);
    break;

  case CMD_ROLL:
    roll_dice(arg);
    break;
  case CMD_DONE:
    turn_done();
    break;
  case CMD_PROPERTY:
    property(arg);
    break;
  case CMD_HOUSES:
    houses(arg);
    break;
  case CMD_JAIL:
    jail(arg);
    break;

  case CMD_TRADE:
    trade(arg);
    break;

  case CMD_PAY:
    pay();
    break;
  case CMD_BANKRUPTCY:
    bankruptcy();
    break;

  case CMD_CONNECT:
    dlg_connect(arg);
    break;
  case CMD_DISCONNECT:
    dlg_disconnect();
    break;

  case CMD_LOG:		/* undocumented */
    freelog(LOG_NORMAL, "%s", arg);
    break;
  case CMD_CRASH:
    crash_and_burn();
    
  case CMD_NUM:
  case CMD_UNRECOGNIZED:
  case CMD_AMBIGUOUS:
  default:
    freelog(LOG_FATAL, "bug in client: impossible command recognized; bye!");
    assert(0);
  }
}

/*
 * copy next token from source string src into destination string dest
 * return position in src, where next token starts (or end of src)
 */
char *InputLine::next_tok(char *src, char *dest)
{
  char *cptr_s, *cptr_d;

  // skip spaces at beginning of src
  for(cptr_s = src; *cptr_s && isspace(*cptr_s); cptr_s++);

  // copy until next space (or end of cstr_s) is read 
  for(cptr_d=dest; *cptr_s && !isspace(*cptr_s); cptr_s++, cptr_d++)
    *cptr_d=*cptr_s;
  *cptr_d='\0';

  // ignore space chars at end of token -> advance to next token
  for(; *cptr_s && isspace(*cptr_s); cptr_s++);

  return cptr_s;
}

/*
 * read an integer from string src
 * return position, where first non alnum char (or end of src) was read
 * errcode:
 * 0: integer was read
 * 1: string was no integer
 * 2: string was empty
 */
char *InputLine::next_tok_int(char *src, int &dest, int &errcode)
{
  char *cptr_s, *endptr[1];

  //skip spaces at beginning of src
  for(cptr_s = src; *cptr_s && isspace(*cptr_s); cptr_s++);

  dest = (int) strtol(cptr_s, endptr, 0);
  
  if (*cptr_s == '\0') { // empty token
    dest = -2;
    errcode = 2;
    return cptr_s;
  } else if (! (isspace(**endptr) || **endptr == '\0') ) { 
    // *endptr doesn't point to space or '\0'
    // -> invalid string
    dest = -1;
    errcode = 1;
    // ignore all characters until a space or '\0' is read
    for(cptr_s = *endptr; *cptr_s && !isspace(*cptr_s); cptr_s++);
  } else {	// OK
    errcode = 0;
    cptr_s = *endptr;
  }
    
  // ignore space chars at end of token
  for(; *cptr_s && isspace(*cptr_s); cptr_s++);

  return cptr_s;

}

int InputLine::lookup_player(char *name)
{
  int i, lastmatch = -1;
    
  for (i=0; i<_view->getGame()->nplayers(); i++) {
    char *plrname = const_cast<char *>(_view->getGame()->player(i)->name());
    if (!mystrncasecmp(name, plrname, strlen(name))) 
      {
	/* If name is a perfect match, not partial, assume
	 * that this is the player the user wanted. */
	if (!mystrcasecmp(name, plrname))
	  return i;
	if (lastmatch != -1)
	  return -2;
	lastmatch = i;
      }
  }
    
  return lastmatch;
}

char *InputLine::next_tok_player(char *str, int &plrid, int &errcode)
{
  char *cptr_s = str, *cptr_f;
  char *fullname;	// full name from game, if found any
  char *oldpos;	// to store position in string
  char parsename[MAX_CMD_LEN+1];	// parsed name
  char buf[MAX_CMD_LEN+1];
  
  //skip spaces at beginning of str
  for(; *str && isspace(*str); str++);

  cptr_s = next_tok(str, parsename);

  if (!*parsename) {	// empty token
    plrid = -3;
    errcode = 3;
    return cptr_s;
  } 

  // test, if there is any player with name "parsename"
  plrid = lookup_player(parsename);

  switch (plrid) {
  case -1:	// parsename is no valid player
    errcode = 1;
    return cptr_s;
  case -2:	// parsename ambigous
    oldpos = cptr_s;
    // is it unique with next tokens ?
    for(;;) {
      if (!*cptr_s) {	// end of str -> still ambigous
	errcode = 2;
	return cptr_s;
      }
      cptr_s = next_tok(cptr_s, buf);	
      strncpy(parsename, str, cptr_s-str);
      parsename[cptr_s-str] = '\0';
      plrid = lookup_player(parsename);
      if (plrid!=-2) break;  // not ambigous any more (either wrong or valid)
    }
    if (plrid==-1) {	// no valid player -> first token was ambigous
      errcode = 2;
      return oldpos;
    } else if (plrid>=0) {	// plrid is OK
      errcode = 0;
      goto remove_rest;
    } else assert(0);	// have we missed any errcode from lookup_player ?
    
  default:	// player OK -> plrid OK

    
    errcode = 0;

  remove_rest:
    // if name was not complete, we must remove all tokens that are part of
    // the name (we allow space in names)

    cptr_f = fullname = 
      const_cast<char *>(_view->getGame()->player(plrid)->name());
    cptr_f = next_tok(cptr_f, buf);
    
    for (;;) {
      oldpos = cptr_s;
      if (!*cptr_s)	// empty next token -> exit
	break;
      // read next tokens in parsed str and in fullname and compare them
      cptr_s = next_tok(cptr_s, parsename);
      cptr_f = next_tok(cptr_f, buf);
      if (mystrncasecmp(parsename, buf, strlen(parsename))) { 
	// difference found -> revert to old position in str
	cptr_s = oldpos;
	break;
      }
    }
    
  }

  return cptr_s;
}


int InputLine::lookup_field(char *name)
{
  int i, lastmatch = -1;
    
  for (i=0; i<BOARD_N_FIELDS; i++) {
    const char *fieldname = _view->field_locale(i)->name;
    if (!mystrncasecmp(name, fieldname, strlen(name))) 
      {
	/* If name is a perfect match, not partial, assume
	 * that this is the field the user wanted. */
	if (!mystrcasecmp(name, fieldname))
	  return i;
	if (lastmatch != -1)
	  return -2;
	lastmatch = i;
      }
  }
    
  return lastmatch;
}

char *InputLine::next_tok_field(char *str, int &fieldid, int &errcode)
{
  char *cptr_s = str, *cptr_f;
  char *fullname;	// full name, if found any
  char *oldpos;	// to store position in string
  char parsename[MAX_CMD_LEN+1];	// parsed name
  char buf[MAX_CMD_LEN+1];
  
  //skip spaces at beginning of str
  for(; *str && isspace(*str); str++);

  cptr_s = next_tok(str, parsename);

  if (!*parsename) {	// empty token
    fieldid = -3;
    errcode = 3;
    return cptr_s;
  } 

  // test, if there is any field with name "parsename"
  fieldid = lookup_field(parsename);

  switch (fieldid) {
  case -1:	// parsename is no valid field
    errcode = 1;
    return cptr_s;
  case -2:	// parsename ambigous
    oldpos = cptr_s;
    // is it unique with next tokens ?
    for(;;) {
      if (!*cptr_s) {	// end of str -> still ambigous
	errcode = 2;
	return cptr_s;
      }
      cptr_s = next_tok(cptr_s, buf);	
      strncpy(parsename, str, cptr_s-str);
      parsename[cptr_s-str] = '\0';
      fieldid = lookup_field(parsename);
      if (fieldid!=-2) break;  // not ambigous any more (either wrong or valid)
    }
    if (fieldid==-1) {	// no valid field -> first token was ambigous
      errcode = 2;
      return oldpos;
    } else if (fieldid>=0) {	// fieldid is OK
      errcode = 0;
      goto remove_rest;
    } else assert(0);	// have we missed any errcode from lookup_field ?
    
  default:	// field OK -> fieldid OK

    errcode = 0;

  remove_rest:
    // if name was not complete, we must remove all tokens that are part of
    // the name (because we allow space in names)

    cptr_f = fullname = const_cast<char *>(_view->field_locale(fieldid)->name);
    cptr_f = next_tok(cptr_f, buf);
    
    for (;;) {
      oldpos = cptr_s;
      if (!*cptr_s)	// empty next token -> exit
	break;
      // read next tokens in parsed str and in fullname and compare them
      cptr_s = next_tok(cptr_s, parsename);
      cptr_f = next_tok(cptr_f, buf);
      if (mystrncasecmp(parsename, buf, strlen(parsename))) { 
	// difference found -> revert to old position in str
	cptr_s = oldpos;
	break;
      }
    }
    
  }

  return cptr_s;
}


int InputLine::lookup_monop(char *name)
{
  int i, lastmatch = -1;
    
  for (i=0; i<BOARD_N_MONOPS; i++) {
    char *monopname = _view->monop_locale(i).color;
    if (!mystrncasecmp(name, monopname, strlen(name))) 
      {
	/* If name is a perfect match, not partial, assume
	 * that this is the monop the user wanted. */
	if (!mystrcasecmp(name, monopname))
	  return i;
	if (lastmatch != -1)
	  return -2;
	lastmatch = i;
      }
  }
    
  return lastmatch;
}

char *InputLine::next_tok_monop(char *str, int &monopid, int &errcode)
{
  char *cptr_s = str, *cptr_f;
  char *fullname;	// full name from game, if found any
  char *oldpos;	// to store position in string
  char parsename[MAX_CMD_LEN+1];	// parsed name
  char buf[MAX_CMD_LEN+1];
  
  //skip spaces at beginning of str
  for(; *str && isspace(*str); str++);

  cptr_s = next_tok(str, parsename);

  if (!*parsename) {	// empty token
    monopid = -3;
    errcode = 3;
    return cptr_s;
  } 

  // test, if there is any monop with name "parsename"
  monopid = lookup_monop(parsename);

  switch (monopid) {
  case -1:	// parsename is no valid monop
    errcode = 1;
    return cptr_s;
  case -2:	// parsename ambigous
    oldpos = cptr_s;
    // is it unique with next tokens ?
    for(;;) {
      if (!*cptr_s) {	// end of str -> still ambigous
	errcode = 2;
	return cptr_s;
      }
      cptr_s = next_tok(cptr_s, buf);	
      strncpy(parsename, str, cptr_s-str);
      parsename[cptr_s-str] = '\0';
      monopid = lookup_monop(parsename);
      if (monopid!=-2) break;  // not ambigous any more (either wrong or valid)
    }
    if (monopid==-1) {	// no valid monop -> first token was ambigous
      errcode = 2;
      return oldpos;
    } else if (monopid>=0) {	// monopid is OK
      errcode = 0;
      goto remove_rest;
    } else assert(0);	// have we missed any errcode from lookup_monop ?
    
  default:	// monop OK -> monopid OK

    
    errcode = 0;

  remove_rest:
    // if name was not complete, we must remove all tokens that are part of
    // the name (we allow space in names)

    cptr_f = fullname = _view->monop_locale(monopid).color;
    cptr_f = next_tok(cptr_f, buf);
    
    for (;;) {
      oldpos = cptr_s;
      if (!*cptr_s)	// empty next token -> exit
	break;
      // read next tokens in parsed str and in fullname and compare them
      cptr_s = next_tok(cptr_s, parsename);
      cptr_f = next_tok(cptr_f, buf);
      if (mystrncasecmp(parsename, buf, strlen(parsename))) { 
	// difference found -> revert to old position in str
	cptr_s = oldpos;
	break;
      }
    }
    
  }

  return cptr_s;
}

  
void InputLine::test_tok_procs()
{
  char arg[MAX_CMD_LEN+1];

  char t[20];
  int ti, err;
  char *tptr;

  // test of next_tok
  strcpy(arg, " h d ful");

  tptr = next_tok(arg, t);
  tptr = next_tok(tptr, t);
  tptr = next_tok(tptr, t);
  tptr = next_tok(tptr, t);
  
  // test of next_tok_int
  strcpy(arg, " 1 -12 12h13 dx fful1234");
  
  tptr = next_tok_int(arg, ti, err);
  tptr = next_tok_int(tptr, ti, err);
  tptr = next_tok_int(tptr, ti, err);
  tptr = next_tok_int(tptr, ti, err);
  tptr = next_tok_int(tptr, ti, err);
  tptr = next_tok_int(tptr, ti, err);

  strcpy(arg, "da hol bob danx hollste dani");
  tptr = next_tok_player(arg, ti, err);
  tptr = next_tok_player(tptr, ti, err);
  tptr = next_tok_player(tptr, ti, err);
  tptr = next_tok_player(tptr, ti, err);
  tptr = next_tok_player(tptr, ti, err);
  tptr = next_tok_player(tptr, ti, err);
}

void InputLine::print_usage(enum command_id cmdid, const char *errstr,
			   const char *usage)
{
  cmd_reply( cmdid, C_SYNTAX, const_cast<char *>(horiz_line) );
  if (errstr) cmd_reply(cmdid, C_SYNTAX, "error: %s", (char *) errstr);
  if (usage) cmd_reply(cmdid, C_SYNTAX, (char *) usage);
  cmd_reply( cmdid, C_SYNTAX, const_cast<char *>(horiz_line) );
}

void InputLine::print_subcommand_list(enum command_id cmdid, int cmdid_last,
				     struct command *cmd)
{ 
  int i;
  cmd_reply( cmdid, C_COMMENT, const_cast<char *>(horiz_line) );
  cmd_reply(cmdid, C_COMMENT,
	    "valid subcommands are:");
  for (i=0; i<cmdid_last; i++) {
    cmd_reply(cmdid, C_COMMENT, "\t%s", cmd[i].name);
  }
  cmd_reply( cmdid, C_COMMENT, const_cast<char *>(horiz_line) );
}





void InputLine::show_help()
{
  cmd_reply(CMD_HELP, C_COMMENT,
	    "Client (Server) commands are prefixed with '.' ('/').");
//    cmd_reply(CMD_HELP, C_COMMENT,
//  	    "All other input is treated as chat.");
  cmd_reply(CMD_HELP, C_COMMENT,
	    "Available commands:\n (type command without arguments to get "
	    "subcommand list)");
  cmd_reply(CMD_HELP, C_COMMENT, const_cast<char *>(horiz_line));
  cmd_reply(CMD_HELP, C_COMMENT,
	    ".help                - this help text");
//    cmd_reply(CMD_HELP, C_COMMENT,
//  	    "list                - list players (with connection)");
//    cmd_reply(CMD_HELP, C_COMMENT,
//  	    "quit                - quit game");
  // player is surely connected
//    cmd_reply(CMD_HELP, C_COMMENT,
//  	    "disconnect          - close connection to server");
  cmd_reply(CMD_HELP, C_COMMENT,
  	    ".chat                - send chat message");

  if (_view->getApp()->client_state() == CLIENT_GAME_RUNNING_STATE) {
    cmd_reply(CMD_HELP, C_COMMENT, const_cast<char *>(horiz_line));
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".info <subcmd>       - information about board, players");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".roll                - roll the dice");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".done                - end your turn");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".property <subcmd>   - manipulate properties (buy etc.)");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".houses <subcmd>     - manipulate houses (buy etc.)");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".trade <subcmd>      - trade with other players");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".jail <subcmd>       - get out of jail");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".pay                 - settle your debts");
    cmd_reply(CMD_HELP, C_COMMENT,
	      ".bankruptcy          - declare yourself bankrupt");
  }
  cmd_reply(CMD_HELP, C_COMMENT, const_cast<char *>(horiz_line));
  cmd_reply(CMD_HELP, C_COMMENT, "Abbreviations are allowed.");
}

void InputLine::show_players()
{
  cmd_reply(CMD_LIST,  C_COMMENT, "List of players:");
  cmd_reply(CMD_LIST,  C_COMMENT, const_cast<char *>(horiz_line));
  cmd_reply(CMD_LIST,  C_COMMENT, "<not implemented yet>");

  /*    if (_view->getGame()->nplayers() == 0) */
  /*      cmd_reply(CMD_LIST,  C_WARNING, "<no players>"); */
  /*    else */
  /*    { */
  /*      for(i=0; i<_view->getGame()->nplayers(); i++) { */
  /*        if (_view->getGame()->player(i)->conn()) { */
  /*  	cmd_reply(CMD_LIST,  C_COMMENT, */
  /*  		  "%s is connected from %s", */
  /*  		  _view->getGame()->player(i)->name(), */
  /*  		  _view->getGame()->player(i)->addr());  */
  /*        } else { */
  /*  	cmd_reply(CMD_LIST,  C_COMMENT, */
  /*  		  "%s is not connected", */
  /*  		  _view->getGame()->player(i)->name()); */
  /*        } */
  /*      } */
  /*    } */
  cmd_reply(CMD_LIST,  C_COMMENT, const_cast<char *>(horiz_line));
}

void InputLine::dlg_quit()  // disabled
{
//    int i = 1;

  cmd_reply(CMD_QUIT, C_FAIL, 
	    "Error: quit not implemented on command line" );
  // TODO: modal dialog to confirm
//    if (i == 1) {
//      _view->getApp()->slotGameQuit();
//    }
}

void InputLine::chat(char *msg)
{
  if (_view->getApp()->client_state() == CLIENT_BOOT_STATE) {	
    // chat not possible yet
    cmd_reply(CMD_CHAT, C_FAIL, "no connection to server yet");
    return;
  }

  if (!*msg) {	// empty message -> usage
    print_usage(CMD_CHAT, 0, "syntax: chat [ playername: ] message");
    return;
  }
  
  char tmpmsg[MSG_SIZE], newmsg[MSG_SIZE];
  chat_process_field_names(msg, tmpmsg);
  chat_process_money(tmpmsg, newmsg);

  struct packet_generic_message pack;
  
  strncpy(pack.message, newmsg, MSG_SIZE);
  pack.message[MSG_SIZE-1]='\0';

  send_packet_generic_message(_view->getApp()->conn(),
			      PACKET_CHAT_MSG, &pack);
}


void InputLine::dlg_connect(char * /*arg*/)
{
//    char errbuf[BUF_SIZE+1];

  cmd_reply(CMD_CONNECT, C_FAIL, 
	    "Error: connect not implemented on command line" );
  // not needed here, TODO: remove

}
  
void InputLine::dlg_disconnect()
{
  // TODO: modal dialog
  cmd_reply(CMD_DISCONNECT, C_FAIL, 
	    "Error: disconnect not implemented on command line" );
//    disconnect_from_server();
}

void InputLine::roll_dice(char *arg)
{
  int errcode;

  char *cptr_s = arg;

  int dice_sum;

  struct packet_generic_integer pack;


  cptr_s = next_tok_int(cptr_s, dice_sum, errcode);

  // if there was an argument and it is and integer, it is taken as dice_sum
  // for cheating
  if (!errcode) { 
    pack.value = dice_sum;
  } else {
    // we ignore any non-existing or wrong argument
    pack.value = -1;
  }

  send_packet_generic_integer(_view->getApp()->conn(),
			      PACKET_DICE_ROLL_REQUEST, &pack);
}

void InputLine::turn_done()
{
  struct packet_generic_integer pack;

  pack.value = _view->getGame()->myPID();

  send_packet_generic_integer(_view->getApp()->conn(),
			      PACKET_TURN_DONE, &pack);
}


void InputLine::info(char *subcmd)
{
  const char *usage =
    "syntax: info <about> [ <name> ]\n"
    "\t<about>:\tboard, field, player\n"
    "\t<name> :\tfield/player name\n";

  char cmdname[MAX_CMD_LEN+1], *cptr_s, *cptr_d;
  int cmd, id, errcode;

  for(cptr_s=subcmd; *cptr_s && !isalnum(*cptr_s); cptr_s++);
  for(cptr_d=cmdname; *cptr_s && isalnum(*cptr_s); cptr_s++, cptr_d++)
    *cptr_d=*cptr_s;
  *cptr_d='\0';

  if (*cmdname) {	/* 2nd argument (subcommand) is given */
    cmd=lookup_subcommand(cmdname, info_cmd, CMD_INFO_LAST);
    if (cmd==-1) {
      cmd_reply(CMD_INFO, C_FAIL, "Unknown subcommand: %s", cmdname);
      return;
    }
    else if (cmd==-2) {
      cmd_reply(CMD_INFO, C_FAIL, "Ambiguous subcommand: %s", cmdname);
      return;
    }
    else {	/* valid subcommand */
      
      switch (cmd) {

      case CMD_INFO_BOARD:
	print_board_info();
	break;

      case CMD_INFO_FIELD:
	// get field name from args, converted to id
	cptr_s = next_tok_field(cptr_s, id, errcode);
	
	if (errcode == 1) {	// no field found with that name
	  print_usage(CMD_INFO, "no field found with that name", usage);
	  return;
	} else if (errcode == 2) {	// field found but ambiguous
	  print_usage(CMD_INFO, "field name is ambiguous", usage);
	  return;
	} else if (errcode == 3) {	// empty token
	  id = -1;	// show current field (location of my player)
	}
	
	// here: id is OK
	print_field_info(id);
	break;

      case CMD_INFO_PLAYER:
	// get player name from args, converted to id
	cptr_s = next_tok_player(cptr_s, id, errcode);
	
	if (errcode == 1) {	// no player found with that name
	  print_usage(CMD_INFO, "no player found with that name", usage);
	  return;
	} else if (errcode == 2) {	// player found but ambiguous
	  print_usage(CMD_INFO, "player name is ambiguous", usage);
	  return;
	} else if (errcode == 3) {	// empty token
	  id = -1;	// show all players
	}
	
	// here: id is OK
	print_player_info(id);
	break;

      default:
	freelog(LOG_FATAL, 
		"bug in client: impossible command recognized in info();\n"
		"bye!");
	assert(0);
      }
    }
  } else {	/* no subcommand given */

    print_usage(CMD_INFO, 0, usage);
  }
}

    

void InputLine::property(char *subcmd)
{
  const char *usage = 
    "syntax: property <subcmd> <propname>\n"
    "\t<subcmd>  :\tbuy, mortgage, unmortgage\n"
    "\t<propname>:\tname of the property\n";

  char cmdname[MAX_CMD_LEN+1], *cptr_s, *cptr_d;
  int cmd, loc, errcode;
  struct packet_generic_integer pack;
  
  for(cptr_s=subcmd; *cptr_s && !isalnum(*cptr_s); cptr_s++);
  for(cptr_d=cmdname; *cptr_s && isalnum(*cptr_s); cptr_s++, cptr_d++)
    *cptr_d=*cptr_s;
  *cptr_d='\0';

  if (*cmdname) {	/* 2nd argument (subcommand) is given */
    cmd=lookup_subcommand(cmdname, prop_cmd, CMD_PROP_LAST);
    if (cmd==-1) {
      cmd_reply(CMD_PROPERTY, C_FAIL, "Unknown subcommand: %s", cmdname);
      return;
    }
    else if (cmd==-2) {
      cmd_reply(CMD_PROPERTY, C_FAIL, "Ambiguous subcommand: %s", cmdname);
      return;
    }
    else {	/* valid subcommand */
      
      // get field name from args, converted to loc
      cptr_s = next_tok_field(cptr_s, loc, errcode);
      
      if (errcode == 1) {	// no field found with that name
	print_usage(CMD_PROPERTY, "no field found with that name", usage);
	return;
      } else if (errcode == 2) {	// field found but ambiguous
	print_usage(CMD_PROPERTY, "field name is ambiguous", usage);
	return;
      } else if (errcode == 3) {	// empty token
	loc = -1;	// show current field (location of my player)
      }
	
      // here: loc is OK
      
      pack.value = (loc!=-1)? loc : _view->getGame()->myPlr()->location();

      switch (cmd) {

      case CMD_PROP_BUY:
	send_packet_generic_integer(_view->getApp()->conn(),
				    PACKET_PROPERTY_BUY, &pack);
	break;
      case CMD_PROP_MORTGAGE:
	send_packet_generic_integer(_view->getApp()->conn(),
				    PACKET_PROPERTY_MORTGAGE, &pack);
	break;
      case CMD_PROP_UNMORTGAGE:
	send_packet_generic_integer(_view->getApp()->conn(),
				    PACKET_PROPERTY_UNMORTGAGE, &pack);
	break;

      default:
	freelog(LOG_FATAL, 
		"bug in client: impossible command recognized in property();\n"
		"bye!");
	assert(0);
      }
    }
  } else {	/* no subcommand given */
    print_usage(CMD_PROPERTY, 0, usage);
//      cmd_reply(CMD_PROPERTY, C_COMMENT, const_cast<char *>(horiz_line));
//      cmd_reply(CMD_PROPERTY, C_COMMENT,
//  	      "Valid subcommands are:");
//      cmd_reply(CMD_PROPERTY, C_COMMENT, const_cast<char *>(horiz_line));
//      for (i=0;i<CMD_PROP_LAST;i++) {
//        cmd_reply(CMD_PROPERTY, C_COMMENT, "%s", prop_cmd[i].name);
//      }
//      cmd_reply(CMD_PROPERTY, C_COMMENT, const_cast<char *>(horiz_line));
  }
}


void InputLine::houses(char *args)
{
  const char *usage =
    "syntax: houses <subcmd> <monopname> <nhouses1> <nhouses2> [<nhouses3>]\n"
    "\t<subcmd>   :\tbuy, sell\n"
    "\t<monopname>:\tmonopoly name (use color name)\n"
    "\t<nhouses1> :\t# houses to buy/sell on 1st street of monopoly etc.";

  char errbuf[200];
  int errcode;

  char *cptr_s;
  char subcmd[MAX_CMD_LEN+1];
  int cmdid;

  int monopid;
  int nhouses[MAX_STREETS_IN_MONOP];

  struct packet_house_request pack;

  int i;

  
  cptr_s = next_tok(args, subcmd);

  if (*subcmd) {	/* 2nd argument (subcommand) is given */
    cmdid=lookup_subcommand(subcmd, houses_cmd, CMD_HOUSES_LAST);
    if (cmdid==-1) {
      cmd_reply(CMD_HOUSES, C_FAIL, "Unknown subcommand: %s", subcmd);
      return;
    }
    else if (cmdid==-2) {
      cmd_reply(CMD_HOUSES, C_FAIL, "Ambiguous subcommand: %s", subcmd);
      return;
    }
    else {	/* valid subcommand */
      
      // get monop name from args, converted to monopid
      cptr_s = next_tok_monop(cptr_s, monopid, errcode);
      
      if (errcode == 1) {	// no monop found with that name
	print_usage(CMD_PROPERTY, "no monop found with that name", usage);
	return;
      } else if (errcode == 2) {	// monop found but ambiguous
	print_usage(CMD_PROPERTY, "monop name is ambiguous", usage);
	return;
      } else if (errcode == 3) {	// empty token
	monopid = -1;	// show current field (location of my player)
      }
	
      // here: monopid is OK
      
      /* get nhouses */
      for (i=0; i<MAX_STREETS_IN_MONOP; ++i) {

	cptr_s = next_tok_int(cptr_s, nhouses[i], errcode);

	if (errcode == 2) {	// empty token
	  if (i < _view->getGame()->board()->monop(monopid)->nstreets()) {
	    // only complain, if there are at least i+1 streets in monop
	    print_usage(CMD_HOUSES, 0, usage);
	    return;
	  }
	} else if (errcode == 1) {	// no integer
	  sprintf(errbuf, "nhouses%d not valid", i+1);
	  print_usage(CMD_HOUSES, errbuf, usage);
	  return;
	} else if (nhouses[i] <0 || nhouses[i] > STREET_MAX_HOUSES+1) {
	  sprintf(errbuf, "nhouses%d out of range: "
		  "0 <= nhouses%d <= %d", i+1, i+1, STREET_MAX_HOUSES + 1);
	  print_usage(CMD_HOUSES, errbuf, usage);
	  return;
	}
      }	
      
      /* now we have the desired numbers, prepare packet */

      pack.monopid = monopid;
      for (i=0; i<MAX_STREETS_IN_MONOP; ++i)
	pack.nhouses[i] = nhouses[i];
      
      switch (cmdid) {
	
      case CMD_HOUSES_BUY:
	send_packet_house_request(_view->getApp()->conn(),
				  PACKET_HOUSES_BUY, &pack);
	break;
      case CMD_HOUSES_SELL:
	send_packet_house_request(_view->getApp()->conn(),
				  PACKET_HOUSES_SELL, &pack);
	break;
	
      default:
	freelog(LOG_FATAL, 
		"bug in client: impossible command recognized in houses();\n"
		"bye!");
	assert(0);
      }
    }
  } else {	/* no subcommand given */
    print_usage(CMD_HOUSES, 0, usage);
//      print_subcommand_list(CMD_HOUSES, CMD_HOUSES_LAST, houses_cmd);
  }
}

void InputLine::jail(char *args)
{
  const char *usage =
    "syntax: jail <subcmd>\n"
    "\t<subcmd>:\tpay, card\n"
    "\tpay the fine, or use jail free card to get out";

  char subcmd[MAX_CMD_LEN+1], *cptr_s;
  int cmdid;

  cptr_s = next_tok(args, subcmd);
  
  if (*subcmd) {	/* 2nd argument (subcommand) is given */
    cmdid = lookup_subcommand(subcmd, jail_cmd, CMD_JAIL_LAST);
    if (cmdid==-1) {
      cmd_reply(CMD_JAIL, C_FAIL, "Unknown subcommand: %s", subcmd);
      return;
    }
    else if (cmdid==-2) {
      cmd_reply(CMD_JAIL, C_FAIL, "Ambiguous subcommand: %s", subcmd);
      return;
    }
    else {	/* valid subcommand */

      struct packet_generic_integer pack;

      switch (cmdid) {

      case CMD_JAIL_PAY:
	pack.value = E_LEAVE_JAIL_PAY_FINE;
	break;
      case CMD_JAIL_CARD:
	pack.value = E_LEAVE_JAIL_USE_CARD;
	break;

      default:
	freelog(LOG_FATAL, 
		"bug in client: impossible command recognized in jail();\n"
		"bye!");
	assert(0);
      }
      
      send_packet_generic_integer(_view->getApp()->conn(),
				  PACKET_JAIL_FREE_REQUEST, &pack);
  
    }
  } else {	/* no subcommand given */

    print_usage(CMD_JAIL, 0, usage);

  }
  

}




void InputLine::trade(char *args)
{
  const char *usage =
    "syntax: trade <subcmd> <destplr> [ <money> <props> ]\n"
    "\t<subcmd>:\toffer, reply, agree, reject\n"
    "\t<destplr>:\tplayer you want to trade with (name)\n"
    "\t<money>:\tmoney to balance the trade\n"
    "\t<props>:\tyour properties to trade *and* these of your counterpart";

  int errcode;

  char *cptr_s;
  char subcmd[MAX_CMD_LEN+1];
  int cmdid;

  int destplr;
  int money;
  int prop[BOARD_N_FIELDS];

  int i;
  int ntok_read, nprops_read, nprop_errors;
  int last_prop;

  struct packet_trade pack;

  Player *myplr = _view->getGame()->myPlr();

  test_tok_procs();


  cptr_s = next_tok(args, subcmd);
  
  if (*subcmd) {	/* 2nd argument (subcommand) is given */
    cmdid = lookup_subcommand(subcmd, trade_cmd, CMD_TRADE_LAST);
    if (cmdid==-1) {
      cmd_reply(CMD_TRADE, C_FAIL, "Unknown subcommand: %s", subcmd);
      return;
    }
    else if (cmdid==-2) {
      cmd_reply(CMD_TRADE, C_FAIL, "Ambiguous subcommand: %s", subcmd);
      return;
    }
    else {	/* valid subcommand */
      
      // get destination player name from args, converted to destplr
      cptr_s = next_tok_player(cptr_s, destplr, errcode);
      
      if (errcode == 1) {	// no player found with that name
	print_usage(CMD_TRADE, "no player found with that name", usage);
	return;
      } else if (errcode == 2) {	// player found but ambiguous
	print_usage(CMD_TRADE, "player name is ambiguous", usage);
	return;
      } else if (errcode == 3) {	// empty token
	print_usage(CMD_TRADE, 0, usage);
	return;
      }
	
      if (destplr == _view->getGame()->myPID()) {
	// you can't trade with yourself
	print_usage(CMD_TRADE, "you can't trade with yourself", usage);
	return;
      }
	
      // here: destplr is OK

      /* get money and properties (for offer and reply) */

      switch (cmdid) {

      case CMD_TRADE_OFFER:
      case CMD_TRADE_REPLY: {

	int mine;
	int yours;
	Trade *trade;
	if (cmdid == CMD_TRADE_OFFER) {
	  mine = T_SRC;
	  yours = T_DEST;
	} else {	// reply -> get info from Trade object 
	  trade = myplr->get_trade_with(_view->getGame()->player(destplr));
	  if (!trade) {	// there is no trade yet
	    print_usage(CMD_TRADE, "no trade to destplr yet, try offer", 
			usage);
	    return;
	  }
	  mine = (trade->get_src() == myplr)? T_SRC : T_DEST;
	  yours = (trade->get_src() == myplr)? T_DEST : T_SRC;
	}

	/* get money */
	cptr_s = next_tok_int(cptr_s, money, errcode);
	
	switch (errcode) {
	case 1:	// invalid token, no integer
	  print_usage(CMD_TRADE, "invalid money", usage); 
	  return;
	case 2:	// empty string
	  print_usage(CMD_TRADE, "money is obligatory for offer/reply", 
		      usage); 
	  return;
	  
	  // test money out of range ???
	}

	// here: money is OK
	// change sign of money, if we are dest
	if (mine == T_DEST) {	// I am dest
	  money = -money;
	}
	
	/* get properties */
	
	for (i=0; i<BOARD_N_FIELDS; ++i)
	  prop[i] = T_NOTRADE;
	
	ntok_read = 0;
	nprops_read = 0; 
	nprop_errors = 0;
	errcode = 0;
	
	
	while (errcode == 0) { // read properties until end of args reached
	  
	  if (nprop_errors > 3) {	// too many errors detected
	    print_usage(CMD_TRADE, "too many errors, ... aborting", usage);
	    return;
	  }
	  
	  // get field name from args, converted to loc
	  cptr_s = next_tok_field(cptr_s, last_prop, errcode);
	  ntok_read++;
      
	  switch (errcode) {
	  case 1: // no field found with that name
	    print_usage(CMD_TRADE, "", usage);
	    cmd_reply(CMD_TRADE, C_SYNTAX, ("invalid prop (arg %d): no field "
					    "with that name (ignored)"),
		      ntok_read);
	    nprop_errors++;
	    continue;
	  case 2: // field found but ambiguous
	    print_usage(CMD_TRADE, "", usage);
	    cmd_reply(CMD_TRADE, C_SYNTAX, ("invalid prop (arg %d): field name"
					    " is ambiguous (ignored)"),
		      ntok_read);
	    nprop_errors++;
	    continue;
	  case 3: // empty token
	    goto loop_end;
	  }
	
	  
	  Property *p = 
	    dynamic_cast<Property *>(_view->getGame()->board()->field(last_prop));

	  if (!p || p->owner() == NULL || 
	      !(p->owner()==myplr || p->owner()->id()==destplr) ) {
	    // is no prop or doesn't belong to me or destplr
	    cmd_reply(CMD_TRADE, C_SYNTAX,
		      "prop (arg %d, id %d) doesn't belong to you or destplr "
		      "(ignored)", ntok_read, last_prop);
	    nprop_errors++;
	    continue;
	  }

	  // here: we have read a valid prop

	  nprops_read++;
	  
	  prop[last_prop] = (p->owner()==myplr)? mine : yours;
	  
	}	// end of read loop
      loop_end:

	// here: all properties read

	// summary: your offer

	// note: summary is to verbose in text field and we have no 
	// confirmation dialog, so skip it on QT-Client
	// TODO: do summary in special trade window

//  	int nprops = 0;
//  	char srcmon[20], destmon[20];
//  	srcmon[0] = destmon[0] = '\0';

//  	cmd_reply(CMD_TRADE, C_COMMENT, "You want to trade with %s:",
//  		  _view->getGame()->player(destplr)->name());
//  	//  	cmd_reply(CMD_TRADE, C_COMMENT, "\t# prop tokens read:\t%d", 
//  	//  		  ntok_read);
//  	//  	cmd_reply(CMD_TRADE, C_COMMENT, "\t# valid properties to trade:\t%d", 
//  	//  		  nprops_read);
//  	//  	cmd_reply(CMD_TRADE, C_COMMENT, "\t# errors:\t%d", nprop_errors);

//  	if (money<0) {	// you would get money
//  	  sprintf(srcmon, "$%d ", -money);
//  	}
//  	if (money>0) {	// dest got money
//  	  sprintf(destmon, "$%d ", money);
//  	}

//  	cmd_reply(CMD_TRADE, C_COMMENT, 
//  		  "  you offer %s%sthe following properties:", 
//  		  destmon, (*destmon)?"and ":"");

//  	for (i=0; i<BOARD_N_FIELDS; ++i) {
//  	  if (prop[i] == mine) {
//  	    cmd_reply(CMD_TRADE, C_COMMENT, "    %-20.20s", 
//  		      _view->field_locale(i)->name);
//  	    nprops++;
//  	  }
//  	}
//  	if (nprops==0)
//  	  cmd_reply(CMD_TRADE, C_COMMENT, "    <no properties>");

//  	// your demand
//  	cmd_reply(CMD_TRADE, C_COMMENT, 
//  		  "  you demand %s%sthe following properties:", 
//  		  srcmon, (*srcmon)?"and ":"");

//  	nprops = 0;
//  	for (i=0; i<BOARD_N_FIELDS; ++i) {
//  	  if (prop[i] == yours) {
//  	    cmd_reply(CMD_TRADE, C_COMMENT, "    %-20.20s", 
//  		      _view->field_locale(i)->name);
//  	    nprops++;
//  	  }
//  	}
//  	if (nprops==0)
//  	  cmd_reply(CMD_TRADE, C_COMMENT, "    <no properties>");
	
//  	// TODO: confirm dialog
//  	bool ok = true;

//  	if (!ok) {
//  	  cmd_reply(CMD_TRADE, C_COMMENT, "...trade aborted");
//  	  return;
//  	}

    	pack.tradeid = 0;	// not used
	pack.event = (cmdid==CMD_TRADE_OFFER)? E_TRADE_OFFER : E_TRADE_REPLY;

  	pack.srcid = _view->getGame()->myPID();	// not used on server side
	pack.destid = destplr;

  	pack.offer_from = T_NOTRADE;	// not used!!!

	pack.money = money;
	

	for (i=0; i<BOARD_N_FIELDS; ++i)
	  pack.prop[i] = prop[i];

	{ // !!! test!!!
//  	  pack.prop[0] = -1;
//  	  pack.prop[1] = T_DEST;
//  	  pack.event = E_TRADE_AGREE;
//  	  pack.destid = 3;
	}

	pack.cards = 0;			// not used yet

	send_packet_trade(_view->getApp()->conn(), &pack);

	// end CMD_TRADE_OFFER
	break;
      }	
      case CMD_TRADE_AGREE:
	pack.event = E_TRADE_AGREE;
	pack.destid = destplr;

	{ // !!! test !!!
//  	  pack.destid = 10;
	}

	send_packet_trade(_view->getApp()->conn(), &pack);
	break;
      case CMD_TRADE_REJECT:
	pack.event = E_TRADE_REJECT;
	pack.destid = destplr;
	send_packet_trade(_view->getApp()->conn(), &pack);
	break;

      default:
	freelog(LOG_FATAL, 
		"bug in client: impossible command recognized in trade();\n"
		"bye!");
	assert(0);
      }
    }      
  } else {	/* no subcommand given */
    
    print_usage(CMD_TRADE, 0, usage);
//      cmd_reply(CMD_TRADE, C_COMMENT, const_cast<char *>(horiz_line));
//      cmd_reply(CMD_TRADE, C_COMMENT,
//  	      "valid subcommands are:");
//      for (i=0;i<CMD_TRADE_LAST;i++) {
//        cmd_reply(CMD_TRADE, C_COMMENT, "\t%s", trade_cmd[i].name);
//      }
//      cmd_reply(CMD_TRADE, C_COMMENT, const_cast<char *>(horiz_line));
  }
  
}


void InputLine::pay()
{
  struct packet_generic_integer pack;

  pack.value = _view->getGame()->myPID();

  send_packet_generic_integer(_view->getApp()->conn(),
			      PACKET_PAY_REQUEST, &pack);
}

void InputLine::bankruptcy()
{
  struct packet_generic_integer pack;

  pack.value = _view->getGame()->myPID();

  send_packet_generic_integer(_view->getApp()->conn(),
			      PACKET_DECLARE_BANKRUPT, &pack);
}


/* for info command */

void InputLine::print_board_info()
{
  int row, lastrow, loc, status, i;
  Field *f;
  Property *p;
  Street *s;
  Railroad *r;
  char buf[80], *oldbufpos;
  int plrloc;
  char lstr[BOARD_N_FIELDS][30];

  cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));

  for (loc=0; loc<BOARD_N_FIELDS; loc++)
    *(lstr[loc]) = '\0';

  for (i=0; i<_view->getGame()->nplayers(); ++i) {
    plrloc = _view->getGame()->player(i)->location();
    sprintf(lstr[plrloc]+strlen(lstr[plrloc]), "%d,", i);
  } 
  // eliminate last comma
  for (loc=0; loc<BOARD_N_FIELDS; ++loc)
    if (*lstr[loc]) *(lstr[loc]+strlen(lstr[loc])-1) = '\0';
  
      
  lastrow = BOARD_N_FIELDS/2 + BOARD_N_FIELDS % 2;

  for (row=0; row<lastrow; ++row) {
    *buf = '\0';
    for (loc=row; loc<BOARD_N_FIELDS; loc += lastrow) {
      f = const_cast<Field *>(_view->getGame()->board()->field(loc));
      p = dynamic_cast<Property *>(f);
      s = dynamic_cast<Street *>(f);
      r = dynamic_cast<Railroad *>(f);

      status = f->status();

      /* location and name */
      sprintf(buf+strlen(buf), "%2d) %-17.17s ", loc, 
	      _view->field_locale(loc)->name);

      oldbufpos = buf + strlen(buf);
      /* substitute name with plrids, if some players are on that field */
      if (*lstr[loc])
	sprintf(oldbufpos-strlen(lstr[loc])-6, " *<%-.9s>* ", lstr[loc]);

      /* owner */
      if (p) {		/* is property */
	if (status & IS_OWNED)
	  sprintf(buf+strlen(buf), "-%d- ", p->owner()->id());
	else
	  sprintf(buf+strlen(buf), "--- ");
      } else
	sprintf(buf+strlen(buf), "    ");
      
      /* monop color */
      if (s) {		/* is street */
	if (status & IS_MONOP)
	  sprintf(buf+strlen(buf), "%c%c ", 
		  _view->monop_locale(s->monop()->id()).color_short,
		  _view->monop_locale(s->monop()->id()).color_short);
	else
	  sprintf(buf+strlen(buf), "%c  ", 
		  _view->monop_locale(s->monop()->id()).color_short);
      } else if (r) {	/* is railroad */
	sprintf(buf+strlen(buf), "   ");
      } else
	sprintf(buf+strlen(buf), "   ");

      /* price */
      if (p) {
	sprintf(buf+strlen(buf), "%5d  ", p->price());
      } else 
	sprintf(buf+strlen(buf), "       ");
      
      /* # of houses (if street and monopoly) */
      if (s  &&  (status & IS_MONOP) )
	sprintf(buf+strlen(buf), "%d  ", s->nhouses()); 
      else
	sprintf(buf+strlen(buf), "   ");

      /* mortgaged ? (overwrite # houses) */
      if (p  && (status & IS_MORTGAGED) )
	sprintf(buf+strlen(buf)-3, "%c  ", 'M');
      
    }
    cmd_reply(CMD_INFO, C_COMMENT, buf);
  }    

  cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));

}

void InputLine::print_field_info(int location)
{
  int loc = (location != -1) ? location:_view->getGame()->myPlr()->location();

  Field *f = const_cast<Field *>(_view->getGame()->board()->field(loc));
  Property *p = dynamic_cast<Property *>(f);
  Street *s = dynamic_cast<Street *>(f);
  Railroad *r = dynamic_cast<Railroad *>(f);
  Utility *u = dynamic_cast<Utility *>(f);


  char buf[6*80];
  int stat;


  cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));

  sprintf(buf, "Field %d: %-20.20s", loc, _view->field_locale(loc)->name);
  if (s)
    sprintf(buf+strlen(buf), "(street, %s)", 
	    _view->monop_locale(s->monop()->id()).color);
  else if (r)
    sprintf(buf+strlen(buf), "(railroad)");
  else if (u)
    sprintf(buf+strlen(buf), "(utility)");

  cmd_reply(CMD_INFO, C_COMMENT, buf);

  if (p) {
    cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));
    sprintf(buf, "  Price      : $%d", p->price());
    cmd_reply(CMD_INFO, C_COMMENT, buf);
    sprintf(buf, "  Mortgage   : $%d", p->price()/2);
    cmd_reply(CMD_INFO, C_COMMENT, buf);
    if (s) {
      sprintf(buf, "  houses cost: $%d", s->house_price());
      cmd_reply(CMD_INFO, C_COMMENT, buf);
    }
    cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));

    // rent
    sprintf(buf, "%-8.8s","  Rent:");

    if (s)
      sprintf(buf+strlen(buf), 
	      "%15.15s:  $%d\n%23.23s:  $%d\n%23.23s:  $%d\n"
	      "%23.23s:  $%d\n%23.23s:  $%d\n%23.23s:  $%d", 
	      "no buildings ", s->rent_for(0),
	      "1 house   ",    s->rent_for(1),
	      "2 houses  ",    s->rent_for(2),
	      "3 houses  ",    s->rent_for(3),
	      "4 houses  ",    s->rent_for(4),
	      "1 hotel   ",    s->rent_for(5) );
    else if (r)
      sprintf(buf+strlen(buf), "$%d * number of railroads owned by owner",
	      r->base_rent());
    else if (u)
      sprintf(buf+strlen(buf), 
	      "%10.10s %-12.12s :   %3d  *  last dice rolled\n"
	      "%18.18s %-12.12s :   %3d  *  last dice rolled",
	      "Owner has", "1 utility", u->rent_factor(0),
	      "", "2 utilities", u->rent_factor(1));
    
    cmd_reply(CMD_INFO, C_COMMENT, buf);
    cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));

    // status
    stat = p->status();

    if (stat & IS_OWNED) {
      // print owner
      sprintf(buf, "  Owner         :  ");
      if (stat & IS_YOURS)
	sprintf(buf+strlen(buf), "YOU");
      else 
	sprintf(buf+strlen(buf), "%s", p->owner()->name());
      if (stat & IS_MONOP)
	sprintf(buf+strlen(buf), " (*** has monopoly ***)");
      cmd_reply(CMD_INFO, C_COMMENT, buf);

      // print other status infos
      if (stat & IS_MORTGAGED) {
	sprintf(buf, "*** this property is MORTGAGED ***");
	cmd_reply(CMD_INFO, C_COMMENT, buf);
      } else {	// not mortgaged
	
	if (s) {
	  sprintf(buf, "  Buildings     :  ");
	  switch (s->nhouses()) {
	  case 0: sprintf(buf+strlen(buf), "--"); break;
	  case 1: sprintf(buf+strlen(buf), "1 house"); break;
	  case 2: case 3: case 4:
	    sprintf(buf+strlen(buf), "%d houses", s->nhouses()); break;
	  case 5:
	    sprintf(buf+strlen(buf), "1 hotel"); break;
	  default:
	    assert(0);	//must be a bug, perhaps other rulesets?
	  }
	  cmd_reply(CMD_INFO, C_COMMENT, buf);
	}      
	sprintf(buf, "  Current Rent  :  $%d", p->rent());
	cmd_reply(CMD_INFO, C_COMMENT, buf);
      }
    } else {	// not owned yet
      sprintf(buf, "*** this property is still FREE ***");
      cmd_reply(CMD_INFO, C_COMMENT, buf);
    }
  }
  cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));
  
}

void InputLine::print_player_info(int plrid)
{
  char buf[6*80];
  char name[PLAYER_MAX_LENGTH_NAME + 10];	// additional "(you)"
  bool me = (plrid == _view->getGame()->myPID());
  Player *plr;
  int i, loc;

  cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));

  for (i=0; i<_view->getGame()->nplayers(); ++i) {
    if (plrid == -1 || i == plrid) {
      
      plr = _view->getGame()->player(i);
      strcpy(name, plr->name());
      me = (i == _view->getGame()->myPID());
      if (me) strcat(name, " (YOU)");
      loc = plr->location();
      
      // id and name
      sprintf(buf, "player %d:  %c %-15.15s ",
	      i, (_view->getGame()->is_on_turn(plr))?'*':' ', name);
      // location
      if (loc == JAIL) {
	int ninjail = plr->nturns_in_jail();
	sprintf(buf+strlen(buf), "in jail (%d turn%s       ", 
		ninjail, (ninjail > 1)?"s)":") " );
      } else
	sprintf(buf+strlen(buf), "on %-20.20s ", 
		_view->field_locale(loc)->name);
	
      // budget
//        if (me) sprintf(buf+strlen(buf), "(budget $%-5d)", plr->budget());
//        else sprintf(buf+strlen(buf), "(budget $???  )");	// budget invisible
      sprintf(buf+strlen(buf), "(has $%-5d)", plr->budget());
      cmd_reply(CMD_INFO, C_COMMENT, buf);
    }
  }

  cmd_reply(CMD_INFO, C_COMMENT, const_cast<char *>(horiz_line));
  
}


  
void InputLine::chat_process_field_names(char *msg, char *newmsg)
{
  char *cptr_s, *cptr_d;
  int loc, errcode;

  cptr_s = msg;
  cptr_d = newmsg;

  while(1) {  // loop until end of msg is detected and return is called

    // find CHAT_FIELDNAME_PREFIX in msg and copy until then
    for(; *cptr_s && (*cptr_s != CHAT_FIELDNAME_PREFIX); cptr_s++, cptr_d++)
      *cptr_d=*cptr_s;

    if (*cptr_s == 0) { // found no more CHAT_FIELDNAME_PREFIX
      *cptr_d='\0';
      return;
    } else {       // found CHAT_FIELDNAME_PREFIX -> process it

      // parse substring after CHAT_FIELDNAME_PREFIX for valid field name
      cptr_s++;
      char *oldpos = cptr_s;
      cptr_s = next_tok_field(cptr_s, loc, errcode);

      if (errcode == 0) {  // field successfully found

	char locstr[6];  // loc can have maximally 3 digits !
	snprintf(locstr, 5, "%c%d ", CHAT_FIELDNAME_PREFIX, loc);
	locstr[5]='\0';
	if (cptr_d-newmsg+1 > MSG_SIZE-1 - strlen(locstr)) {
	  cmd_reply(CMD_CHAT, C_WARNING, ("reached MSG_SIZE limit, "
					  "not processing field name") );
	} else {
	  sprintf(cptr_d, "%s", locstr);
	  cptr_d += strlen(locstr);
	}
      
      } else {  // some problem with the field lookup
	if (errcode == 1) {	// no field found with that name
	  cmd_reply(CMD_CHAT, C_WARNING,
		    ("warning: field name in chat msg not found") );
	} else if (errcode == 2) {	// field found but ambiguous
	  cmd_reply(CMD_CHAT, C_WARNING, 
		    ("warning: field name in chat msg is ambiguous") );
	} else if (errcode == 3) {	// empty token
	  // ignore CHAT_FIELDNAME_PREFIX and continue
	}
	cptr_s = oldpos;  // continue just after the CHAT_FIELDNAME_PREFIX
      }

    }
    
  }
}	


void InputLine::chat_process_money(char *msg, char *newmsg)
{
  char *cptr_s, *cptr_d;
  int money, errcode;

  cptr_s = msg;
  cptr_d = newmsg;

  while(1) {  // loop until end of msg is detected and return is called

    // find CHAT_MONEY_PREFIX in msg and copy until then
    for(; *cptr_s && (*cptr_s != CHAT_MONEY_PREFIX); cptr_s++, cptr_d++)
      *cptr_d=*cptr_s;

    if (*cptr_s == 0) { // found no more CHAT_MONEY_PREFIX
      *cptr_d='\0';
      return;
    } else {       // found CHAT_MONEY_PREFIX -> process it

      // parse substring after CHAT_MONEY_PREFIX for valid integer
      cptr_s++;
      char *oldpos = cptr_s;

      cptr_s = next_tok_int(cptr_s, money, errcode);

      if (errcode == 0) {  // money successfully extracted
	
	money = (int) round( (double) money /
			     _view->money_locale()->factor );
	char moneystr[100];  // hopefully enough room for money string
	snprintf(moneystr, 100, "%c%d ", CHAT_MONEY_PREFIX, money);
	moneystr[99]='\0';
	if (cptr_d-newmsg+1 > MSG_SIZE-1 - strlen(moneystr)) {
	  cmd_reply(CMD_CHAT, C_WARNING, ("reached MSG_SIZE limit, "
					  "not processing money") );
	} else {
	  sprintf(cptr_d, "%s", moneystr);
	  cptr_d += strlen(moneystr);
	}
      
	
      } else {  // some problem with the money string

	if (errcode == 1) {	// no integer
	  cmd_reply(CMD_CHAT, C_WARNING,
		    ("warning: money string in chat msg is no integer") );
	} else if (errcode == 2) {	// empty token
	  // ignore CHAT_FIELDNAME_PREFIX and continue
	}
	cptr_s = oldpos;  // continue just after the CHAT_MONEY_PREFIX
      }

    }
    
  }
}



/*****************************************************/

UpDownFrame::UpDownFrame(KapView *view, QWidget *parent)
  : QFrame(parent), _view(view)
{
  setBackgroundColor(parent->backgroundColor());
  _history = new QStrList;
  _history.append("");	// Last Entry should always be the empty string !
}
 
UpDownFrame::~UpDownFrame()
{
}

void UpDownFrame::set_inputLine(InputLine *input)
{
  _input = input;
}

void UpDownFrame::append_history(const char *line)
{
  _history.removeLast();
  _history.append(line);
  _history.append("");
}

void UpDownFrame::keyPressEvent( QKeyEvent *e )
{
  char *txt = 0;

  assert(_input != NULL);	// _input should be set by MainWindow

  switch( e->key() ) {                        // Look at the key code
  case Key_Up:
    txt = _history.prev();
    if (!txt)	// we were at beginning of list
      txt = _history.first();
    break;
  case Key_Down:
    txt = _history.next();
    if (!txt)	// we were at beginning of list
      _history.last();
    break;
  default:                                // If not an interesting key,
    e->ignore();                        // we don't accept the event
    return;  
  }
  if (txt) _input->setText(txt);

}
