// -*- C++ -*-

/***************************************************************************
                          inputline.h  -  description
                             -------------------
    begin                : Mon Jun 11 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef INPUTLINE_H
#define INPUTLINE_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif 


//
// Qt
//

#include <qlineedit.h>
#include <qframe.h>
#include <qstrlist.h>

class QBoxLayout;

//
// View
//

class KapView;
class UpDownFrame;



#define SERVER_COMMAND_PREFIX '/'
#define CLIENT_COMMAND_PREFIX '.'

/**
 *
 */
class InputLine : public QLineEdit {
  Q_OBJECT

public:

  enum {
    CHAT_MODE = 0,
    GAME_MODE = 1,
    LAST_MODE = 2
  };

  /**
   * Order here is important: for ambiguous abbreviations the first
   * match is used.  Arrange order to:
   * - allow most common short form commands (like 'd' for done) to work.
   * - reduce harm for ambiguous cases, where "harm" includes inconvenience,
   * eg accidently quitting the game.
   */
  enum command_id {
    /* one-letter commands: */
    CMD_HELP = 0,
    CMD_LIST,
    CMD_QUIT,
    CMD_CHAT,
    
    CMD_INFO,
    
    CMD_ROLL,
    CMD_DONE,
    CMD_PROPERTY,
    CMD_HOUSES,
    CMD_JAIL,
    CMD_TRADE,

    CMD_PAY,
    CMD_BANKRUPTCY,
    
    CMD_CONNECT,
    CMD_DISCONNECT,
    
    /* undocumented */
    CMD_LOG,
    CMD_CRASH,
    
    /* pseudo-commands: */
    CMD_NUM,		/* the number of commands - for iterations */
    CMD_UNRECOGNIZED,	/* used as a possible iteration result */
    CMD_AMBIGUOUS		/* used as a possible iteration result */
    
  };
  
  enum info_cmd_id {
    CMD_INFO_BOARD,
    CMD_INFO_FIELD,
    CMD_INFO_PLAYER,
    CMD_INFO_LAST
  };
  
  enum prop_cmd_id {
    CMD_PROP_BUY = 0,
    CMD_PROP_MORTGAGE,
    CMD_PROP_UNMORTGAGE,
    CMD_PROP_LAST
  };
  
  enum houses_cmd_id {
    CMD_HOUSES_BUY = 0,
    CMD_HOUSES_SELL,
    CMD_HOUSES_LAST
  };
  
  enum jail_cmd_id {
    CMD_JAIL_PAY = 0,
    CMD_JAIL_CARD,
    CMD_JAIL_LAST
  };
  
  enum trade_cmd_id {
    CMD_TRADE_OFFER = 0,
    CMD_TRADE_REPLY,
    CMD_TRADE_AGREE,
    CMD_TRADE_REJECT,
    CMD_TRADE_LAST
  };
  

public:

  InputLine(KapView *view, QWidget *parent, int mode = CHAT_MODE);
  virtual ~InputLine();

  void set_frame(UpDownFrame *);

public slots:

  void handle_line_input();

protected:

  KapView *_view;

  UpDownFrame *_frm;
  QBoxLayout *_top;

  int _mode;

private:

  static const char *horiz_line;


  /**
   * Commands - can be recognised by unique prefix
   */
  struct command {
    const char *name;              /* name - will be matched by unique prefix   */
  };

static const command commands[];
static const command info_cmd[];
static const command prop_cmd[];
static const command houses_cmd[];
static const command jail_cmd[];
static const command trade_cmd[];



private:

  /**
   *  Convert a named command into an id.
   *  If accept_ambiguity is true, return the first command in the
   *  enum list which matches, else return CMD_AMBIGOUS on ambiguity.
   *  (This is a trick to allow ambiguity to be handled in a flexible way
   *  without importing notify_player() messages inside this routine - rp)
   */
  enum command_id command_named(char *token, int accept_ambiguity);
  /**
   * Find subcommand index by name. Return index (>=0) on success, -1 if no
   * suitable command were found, -2 if several matches were found.
   */
  int lookup_subcommand(char *find, const struct command cmd[], int cmd_last);
  /**
   * feedback related to server commands
   * caller == NULL means console.
   * No longer duplicate all output to console.
   * 'console_id' should be one of the C_ identifiers in console.h
   */
  void cmd_reply(enum command_id cmd, int console_id, const char *format, ...);

  void crash_and_burn();

  // TODO: include it in handle_line_input
  void handle_stdin_input(char *str);

  /**
   * copy next token from source string src into destination string dest
   * return position in src, where next token starts (or end of src)
   */
  char *next_tok(char *src, char *dest);

  /*
   * read an integer from string src
   * return position, where first non alnum char (or end of src) was read
   * errcode:
   * 0: integer was read
   * 1: string was no integer
   * 2: string was empty
   */
  char *next_tok_int(char *src, int &dest, int &errcode);

  /**
   * Find player index by name. Return index (>=0) on success, -1 if no
   * suitable player were found, -2 if several matches were found.
   */
  int lookup_player(char *name);

  char *next_tok_player(char *str, int &plrid, int &errcode);

  /**
   * Find field index by name. Return index (>=0) on success, -1 if no
   * suitable field were found, -2 if several matches were found.
   */
  int lookup_field(char *name);

  char *next_tok_field(char *str, int &fieldid, int &errcode);

  /**
   * Find monop index by name. Return index (>=0) on success, -1 if no
   * suitable monop were found, -2 if several matches were found.
   */
  int lookup_monop(char *name);

  char *next_tok_monop(char *str, int &monopid, int &errcode);

  void test_tok_procs();
  void print_usage(enum command_id cmdid, const char *errstr, 
		   const char *usage);
  void print_subcommand_list(enum command_id cmdid, int cmdid_last,
			     struct command *cmd);


  /**
   * more internal functions
   */
  void show_help();
  void show_players();
  void dlg_quit();
  void chat(char *msg);
  
  void info(char *arg);
  
  void roll_dice(char *arg);
  void turn_done();
  void property(char *arg);
  void houses(char *arg);
  void jail(char *arg);
  void trade(char *arg);
  
  void pay();
  void bankruptcy();

  void dlg_connect(char *arg);
  void dlg_disconnect();
  
  /* for info command */
  void print_board_info();
  void print_field_info(int loc);
  void print_player_info(int id);

  /**
   * process field name shortcuts in chat message
   */
  void chat_process_field_names(char *msg, char *newmsg);

  /**
   * process money occurences in chat message
   */
  void chat_process_money(char *msg, char *newmsg);

};

class UpDownFrame : public QFrame {
  Q_OBJECT

public:
  UpDownFrame(KapView *view, QWidget *parent);
  virtual ~UpDownFrame();

  void set_inputLine(InputLine *input);
  void append_history(const char *line);

protected:
  
  KapView *_view;

  InputLine *_input;
  /*
   * history (doubly linked list of char*)
   */
  QStrList _history;
  
protected:

  virtual void keyPressEvent( QKeyEvent *e );

};



#endif //  INPUTLINE_H








