// -*- C++ -*-

/***************************************************************************
                          playerview.cpp  -  description
                             -------------------
    begin                : Mon Jun 11 2001
    copyright            : (C) 2001 by Daniel Hermann
    email                : Daniel.Hermann@stud.uni-karlsruhe.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

//
// Qt
//

#include <qlabel.h>
#include <qpixmap.h>
#include <qbitmap.h>
#include <qtimer.h>

//
// App
//

#include "events.h"

//
// Game
//

#include "player.h"

//
// View
//

#include "kapview.h"
#include "boardview.h"
#include "fieldview.h"
#include "playerview.h"


PlayerView::PlayerView(KapView *view, QWidget *parent, Player *plr)
  : QLabel(parent), _view(view),
    _plr(plr),
    _curr_loc(-1),	// must be updated once
    _curr_target(-1),
    _time_moveOneStep(MEDIUM_SPEED),
    _moveTimerActive(false),
    _animationEnabled(false),
    _movingDirection(DIRECTION_FORWARD),
    _moveTimer(this)
{
  assert(plr);

  _tokenColor = token_color[plr->id()];
  setBackgroundColor(_tokenColor);
  setAlignment(AlignCenter);
  resize(20,20);
  setMask(_view->getBitmap("plrspritemask"));
  setPixmap(_view->getPixmap("plrsprite"));
  
  connect( &_moveTimer, SIGNAL(timeout()), SLOT(slotMoveOn()) );
//    hide();
}

PlayerView::~PlayerView()
{
}

QLabel *PlayerView::get_token(QWidget *use_parent)
{
  QLabel *l = new QLabel(use_parent);
  l->setBackgroundColor(_tokenColor);
  l->setAlignment(AlignCenter);
//    l->resize(20,20);
  l->setMask(_view->getBitmap("plrspritemask"));
  l->setPixmap(_view->getPixmap("plrsprite"));
  return l;
}




void PlayerView::update_status(int event)
{
  int new_loc = _plr->location();
  int direction = ( (event == E_MOVE_BACKWARDS) 
		    ? DIRECTION_BACKWARD : DIRECTION_FORWARD );
  // remove the token when the player is dead
  if (!_plr->is_alive()) hide();

  if (_curr_loc == -1) {		// position not initialized yet
    _view->board()->field(new_loc)->enter(this);
    _curr_loc = new_loc;
    show();
  } else if (new_loc != _curr_loc) {	// position has changed
    move_to(new_loc, direction);
    show();
  }
}


void PlayerView::move_to(const int new_loc, const int direction)
{
  int distance;

  if (new_loc == _curr_loc) return;

  _movingDirection = direction;

  // moving directly to JAIL
  // moving directly from JAIL to JUST_VISITING
  if ( new_loc == JAIL ||
       (_curr_loc == JAIL && new_loc == JUST_VISITING) ) {
    moveDirectly(_curr_loc, new_loc);
    _curr_target = new_loc; // paranoia
    return;
  }

  // otherwise lets fix only the target (which can safely change during the
  // animation, if necessary)
  _curr_target = new_loc;

  // get distance to target field
  distance = _curr_target - _curr_loc;
  if (direction == DIRECTION_FORWARD && distance < 0) // we pass go  
    distance = BOARD_N_FIELDS + distance;
  if (direction == DIRECTION_BACKWARD && distance > 0) // -"- backwards
    distance = BOARD_N_FIELDS - distance;
  
  distance = ( (distance > 0) ? distance : -distance );  // = abs(distance)

  // set speed according to distance to target
  if (distance > 12) _time_moveOneStep = FAST_SPEED;
  else if (distance > 4) _time_moveOneStep = MEDIUM_SPEED;
  else _time_moveOneStep = SLOW_SPEED;

  // now move on until _curr_target is reached
  if (!_moveTimerActive) {
    _view->getApp()->animationStarted();
    slotMoveOn();
  }

}

//  void PlayerView::enableAnimation(const bool anim)
//  {
//    _animationEnabled = anim;
//    if (_animationEnabled) _view->getApp()->animationStarted();
//    else _view->getApp()->animationFinished();
//  }

void PlayerView::moveOneStep()
{
  int new_loc;

  if (_movingDirection == DIRECTION_FORWARD) {
    new_loc = (_curr_loc+1) % BOARD_N_FIELDS;
  } else {
    new_loc = (_curr_loc-1); 
    // new_loc could be < 0 !
    if (new_loc < 0) new_loc = 0;
  }

  moveDirectly(_curr_loc, new_loc);
}

void PlayerView::moveDirectly(const int src_loc, const int dest_loc)
{
  FieldView *old_field;
  FieldView *new_field;

  // first leave old location (if there was any)
  if (src_loc != -1) {
    old_field = _view->board()->field(src_loc);
    old_field->leave(this);
  }

  // then we can enter the new one
  new_field = _view->board()->field(dest_loc);
  new_field->enter(this);

  _curr_loc = dest_loc;

  show();
}

// TODO: replace _moveTimerActive by _moveTimer.isActive()
void PlayerView::slotMoveOn()
{
  // every time we get here, a timer timed out (except for first call)
  _moveTimerActive = false;
  if (_curr_loc != _curr_target) {
    moveOneStep();
    // move on after a pause
    _moveTimer.start( _time_moveOneStep, TRUE );
    _moveTimerActive = true;
  } else { // we've reached the target field
    _view->getApp()->animationFinished();
  }
}

QLabel *PlayerView::get_default_token(KapView *use_view,
				      QWidget *use_parent)
{
  QLabel *l = new QLabel(use_parent);
  l->setAlignment(AlignCenter);
  l->resize(20,20);
  l->setMask(use_view->getBitmap("plrspritemask"));
  l->setPixmap(use_view->getPixmap("plrsprite"));
  return l;
}
